from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Unicode

from flexget import db_schema, plugin
from flexget.db_schema import Session
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import qualities
from flexget.utils.database import quality_property
from flexget.utils.tools import group_entries, parse_timedelta

logger = logger.bind(name='timeframe')

Base = db_schema.versioned_base('upgrade', 0)

entry_actions = {'accept': Entry.accept, 'reject': Entry.reject}


class EntryTimeFrame(Base):
    __tablename__ = 'timeframe'

    id = Column(Unicode, primary_key=True, index=True)
    status = Column(Unicode)
    title = Column(Unicode)
    _quality = Column('quality', String)
    quality = quality_property('_quality')
    first_seen = Column(DateTime, default=datetime.now())
    proper_count = Column(Integer, default=0)

    def __str__(self):
        return '<Timeframe(id=%s,added=%s,quality=%s)>' % (self.id, self.added, self.quality)


class FilterTimeFrame:
    schema = {
        'type': 'object',
        'properties': {
            'identified_by': {'type': 'string', 'default': 'auto'},
            'target': {'type': 'string', 'format': 'quality_requirements'},
            'wait': {'type': 'string', 'format': 'interval'},
            'on_waiting': {
                'type': 'string',
                'enum': ['accept', 'reject', 'do_nothing'],
                'default': 'reject',
            },
            'on_reached': {
                'type': 'string',
                'enum': ['accept', 'reject', 'do_nothing'],
                'default': 'do_nothing',
            },
        },
        'required': ['target', 'wait'],
        'additionalProperties': False,
    }

    # Run last so we work on only accepted entries
    @plugin.priority(plugin.PRIORITY_LAST)
    def on_task_filter(self, task, config):
        if not config:
            return

        identified_by = (
            '{{ media_id }}' if config['identified_by'] == 'auto' else config['identified_by']
        )

        grouped_entries = group_entries(task.accepted, identified_by)
        if not grouped_entries:
            return

        action_on_waiting = (
            entry_actions[config['on_waiting']] if config['on_waiting'] != 'do_nothing' else None
        )
        action_on_reached = (
            entry_actions[config['on_reached']] if config['on_reached'] != 'do_nothing' else None
        )

        with Session() as session:
            # Prefetch Data
            existing_ids = (
                session.query(EntryTimeFrame)
                .filter(EntryTimeFrame.id.in_(grouped_entries.keys()))
                .all()
            )
            existing_ids = {e.id: e for e in existing_ids}

            for identifier, entries in grouped_entries.items():
                if not entries:
                    continue

                id_timeframe = existing_ids.get(identifier)
                if not id_timeframe:
                    id_timeframe = EntryTimeFrame()
                    id_timeframe.id = identifier
                    id_timeframe.status = 'waiting'
                    id_timeframe.first_seen = datetime.now()
                    session.add(id_timeframe)

                if id_timeframe.status == 'accepted':
                    logger.debug(
                        'Previously accepted {} with {} skipping', identifier, id_timeframe.title
                    )
                    continue

                # Sort entities in order of quality and best proper
                entries.sort(key=lambda e: (e['quality'], e.get('proper_count', 0)), reverse=True)
                best_entry = entries[0]

                logger.debug(
                    'Current best for identifier {} is {}', identifier, best_entry['title']
                )

                id_timeframe.title = best_entry['title']
                id_timeframe.quality = best_entry['quality']
                id_timeframe.proper_count = best_entry.get('proper_count', 0)

                # Check we hit target or better
                target_requirement = qualities.Requirements(config['target'])
                target_quality = qualities.Quality(config['target'])
                if (
                    target_requirement.allows(best_entry['quality'])
                    or best_entry['quality'] >= target_quality
                ):
                    logger.debug(
                        'timeframe reach target quality {} or higher for {}',
                        target_quality,
                        identifier,
                    )
                    if action_on_reached:
                        action_on_reached(best_entry, 'timeframe reached target quality or higher')
                    continue

                # Check if passed wait time
                expires = id_timeframe.first_seen + parse_timedelta(config['wait'])
                if expires <= datetime.now():
                    logger.debug(
                        'timeframe expired, releasing quality restriction for {}', identifier
                    )
                    if action_on_reached:
                        action_on_reached(best_entry, 'timeframe wait expired')
                    continue

                # Verbose waiting, add to backlog
                if action_on_waiting:
                    for entry in entries:
                        action_on_waiting(entry, 'timeframe waiting')
                diff = expires - datetime.now()
                hours, remainder = divmod(diff.seconds, 3600)
                hours += diff.days * 24
                minutes, _ = divmod(remainder, 60)

                logger.info(
                    '`{}`: timeframe waiting for {:02d}h:{:02d}min. Currently best is `{}`.',
                    identifier,
                    hours,
                    minutes,
                    best_entry['title'],
                )

                # add best entry to backlog (backlog is able to handle duplicate adds)
                plugin.get('backlog', self).add_backlog(task, best_entry, session=session)

    def on_task_learn(self, task, config):
        if not config:
            return

        identified_by = (
            '{{ media_id }}' if config['identified_by'] == 'auto' else config['identified_by']
        )

        grouped_entries = group_entries(task.accepted, identified_by)
        if not grouped_entries:
            return

        with Session() as session:
            # Prefetch Data
            existing_ids = (
                session.query(EntryTimeFrame)
                .filter(EntryTimeFrame.id.in_(grouped_entries.keys()))
                .all()
            )
            existing_ids = {e.id: e for e in existing_ids}

            for identifier, entries in grouped_entries.items():
                if not entries:
                    continue

                id_timeframe = existing_ids.get(identifier)

                if not id_timeframe:
                    continue

                # Sort entities in order of quality
                entries.sort(key=lambda e: e['quality'], reverse=True)

                # First entry will be the best quality
                best_entry = entries[0]

                id_timeframe.quality = best_entry['quality']
                id_timeframe.title = best_entry['title']
                id_timeframe.proper_count = best_entry.get('proper_count', 0)
                id_timeframe.status = 'accepted'


@event('plugin.register')
def register_plugin():
    plugin.register(FilterTimeFrame, 'timeframe', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='manual')


class ManualTask:
    """Only execute task when specified with --tasks"""

    schema = {'type': 'boolean'}

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_start(self, task, config):
        # Make sure we need to run
        if not config:
            return
        # If --task hasn't been specified disable this plugin
        # TODO: allow_manual is confusing. Make it less confusing.
        if (
            not task.options.tasks
            or task.name not in task.options.tasks
            or not task.options.allow_manual
        ):
            logger.debug(
                'Disabling task {}, task can only run in manual mode (via API/CLI)', task.name
            )
            task.abort('manual task not specified in --tasks', silent=True)


@event('plugin.register')
def register_plugin():
    plugin.register(ManualTask, 'manual', api_ver=2)



import inspect

from flexget import options
from flexget.event import event
from flexget.terminal import TerminalTable, TerminalTableError, console, table_parser
from flexget.utils.template import get_filters


def do_cli(manager, options):
    header = ['Name', 'Description']
    table_data = [header]
    for filter_name, filter in get_filters().items():
        if options.name and not options.name in filter_name:
            continue
        filter_doc = inspect.getdoc(filter) or ''
        table_data.append([filter_name, filter_doc])
    try:
        table = TerminalTable(options.table_type, table_data)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))
    else:
        console(table.output)


@event('options.register')
def register_parser_arguments():
    # Register subcommand
    parser = options.register_command(
        'jinja-filters',
        do_cli,
        help='View registered jinja2 filters and their description',
        parents=[table_parser],
    )
    parser.add_argument('--name', help='Filter results by filter name')



from flexget import options
from flexget.event import event
from flexget.terminal import console
from flexget.utils.database import with_session
from flexget.webserver import WeakPassword, change_password, generate_token, get_user


@with_session
def do_cli(manager, options, session=None):
    if options.action == 'passwd':
        try:
            change_password(password=options.password, session=session)
        except WeakPassword as e:
            console(e.value)
            return
        console('Updated password')

    if options.action == 'gentoken':
        token = generate_token(session=session)
        console('Generated new token %s' % token)

    if options.action == 'showtoken':
        user = get_user()
        console('Token: %s' % user.token)


@event('options.register')
def register_parser_arguments():
    parser = options.register_command('web', do_cli, help='Manage web server settings')
    subparsers = parser.add_subparsers(dest='action', metavar='<action>')

    pwd_parser = subparsers.add_parser('passwd', help='change password for web server')
    pwd_parser.add_argument('password', metavar='<new password>', help='New Password')

    subparsers.add_parser('gentoken', help='Generate a new api token')
    subparsers.add_parser('showtoken', help='Show api token')



import getpass
import smtplib
import socket
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from smtplib import SMTPAuthenticationError, SMTPSenderRefused, SMTPServerDisconnected

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning

plugin_name = 'email'
logger = logger.bind(name=plugin_name)


class EmailNotifier:
    """
    Send an e-mail with the list of all succeeded (downloaded) entries.

    Configuration options

    ===============  ===================================================================
    Option           Description
    ===============  ===================================================================
    from             The email address from which the email will be sent (required)
    to               The email address of the recipient (required)
    smtp_host        The host of the smtp server
    smtp_port        The port of the smtp server
    smtp_username    The username to use to connect to the smtp server
    smtp_password    The password to use to connect to the smtp server
    smtp_tls         Should we use TLS to connect to the smtp server
    smtp_ssl         Should we use SSL to connect to the smtp server
    ===============  ===================================================================

    Config basic example::

      notify:
        entries:
          via:
            - email:
                from: xxx@xxx.xxx
                to: xxx@xxx.xxx
                smtp_host: smtp.host.com

    Config example with smtp login::

      notify:
        entries:
          via:
            - email:
                from: xxx@xxx.xxx
                to: xxx@xxx.xxx
                smtp_host: smtp.host.com
                smtp_port: 25
                smtp_login: true
                smtp_username: my_smtp_login
                smtp_password: my_smtp_password
                smtp_tls: true

    GMAIL example::

      notify:
        entries:
          via:
            - email:
                from: from@gmail.com
                to: to@gmail.com
                smtp_host: smtp.gmail.com
                smtp_port: 587
                smtp_login: true
                smtp_username: gmailUser
                smtp_password: gmailPassword
                smtp_tls: true

    Default values for the config elements::

      notify:
        entries:
          via:
            - email:
                smtp_host: localhost
                smtp_port: 25
                smtp_login: False
                smtp_username:
                smtp_password:
                smtp_tls: False
                smtp_ssl: False
    """

    def __init__(self):
        self.mail_server = None
        self.host = None
        self.port = None
        self.username = None
        self.password = None
        self.ssl = None
        self.tls = None

    def connect_to_smtp_server(self, config):
        self.host = config['smtp_host']
        self.port = config['smtp_port']
        self.ssl = config['smtp_ssl']
        self.tls = config['smtp_tls']
        self.username = config.get('smtp_username')
        self.password = config.get('smtp_password')
        try:
            logger.debug('connecting to smtp server {}:{}', self.host, self.port)
            self.mail_server = smtplib.SMTP_SSL if self.ssl else smtplib.SMTP
            self.mail_server = self.mail_server(self.host, self.port)
            if self.tls:
                self.mail_server.ehlo()
                self.mail_server.starttls()
                self.mail_server.ehlo()
        except (socket.error, OSError) as e:
            raise PluginWarning(str(e))

        try:
            if self.username:
                # Forcing to use `str` type
                logger.debug('logging in to smtp server using username: {}', self.username)
                self.mail_server.login(self.username, self.password)
        except (OSError, SMTPAuthenticationError) as e:
            raise PluginWarning(str(e))

    schema = {
        'type': 'object',
        'properties': {
            'to': one_or_more({'type': 'string', 'format': 'email'}),
            'from': {
                'type': 'string',
                'default': 'flexget_notifer@flexget.com',
                'format': 'email',
            },
            'autofrom': {'type': 'boolean', 'default': False},
            'smtp_host': {'type': 'string', 'default': 'localhost'},
            'smtp_port': {'type': 'integer', 'default': 25},
            'smtp_username': {'type': 'string'},
            'smtp_password': {'type': 'string'},
            'smtp_tls': {'type': 'boolean', 'default': False},
            'smtp_ssl': {'type': 'boolean', 'default': False},
            'html': {'type': 'boolean', 'default': False},
        },
        'required': ['to'],
        'dependencies': {
            'smtp_username': ['smtp_password'],
            'smtp_password': ['smtp_username'],
            'smtp_ssl': ['smtp_tls'],
        },
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send an email notification

        :param str message: message body
        :param str title: message subject
        :param dict config: email plugin config
        """

        if not isinstance(config['to'], list):
            config['to'] = [config['to']]

        email = MIMEMultipart('alternative')
        email['To'] = ','.join(config['to'])
        email['From'] = (
            getpass.getuser() + '@' + socket.getfqdn() if config['autofrom'] else config['from']
        )
        email['Subject'] = title
        email['Date'] = formatdate(localtime=True)
        content_type = 'html' if config['html'] else 'plain'
        email.attach(MIMEText(message.encode('utf-8'), content_type, _charset='utf-8'))

        # Making sure mail server connection will remain open per host or username
        # (in case several mail servers are used in the same task)
        if not self.mail_server or not (
            self.host == config['smtp_host'] and self.username == config.get('smtp_username')
        ):
            self.connect_to_smtp_server(config)

        connection_error = None
        while True:
            try:
                self.mail_server.sendmail(email['From'], config['to'], email.as_string())
                break
            except (SMTPServerDisconnected, SMTPSenderRefused) as e:
                if not connection_error:
                    self.connect_to_smtp_server(config)
                    connection_error = e
                else:
                    raise PluginWarning('Could not connect to SMTP server: %s' % str(e))


@event('plugin.register')
def register_plugin():
    plugin.register(EmailNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger

from flexget import options, plugin
from flexget.config_schema import register_config_key
from flexget.event import event
from flexget.utils.tools import MergeException

plugin_name = 'template'
logger = logger.bind(name=plugin_name)


class PluginTemplate:
    """
    Appyly templates with preconfigured plugins to a task config.

    Example::

      template: movies

    Example, list of templates::

      template:
        - movies
        - imdb
    """

    schema = {
        'oneOf': [
            {
                'description': 'Apply multiple templates to this task.',
                'type': 'array',
                'items': {'$ref': '#/definitions/template'},
            },
            {
                'description': 'Apply a single template to this task.',
                'allOf': [{'$ref': '#/definitions/template'}],
            },
            {
                'description': 'Disable all templates on this task.',
                'type': 'boolean',
                'enum': [False],
            },
        ],
        'definitions': {
            'template': {
                'type': 'string',
                'description': 'Name of a template which will be applied to this task.',
                'links': [{'rel': 'settings', 'href': '/api/config/templates/{$}'}],
            }
        },
    }

    def prepare_config(self, config):
        if config is None or isinstance(config, bool):
            config = []
        elif isinstance(config, str):
            config = [config]
        return config

    @plugin.priority(257)
    def on_task_prepare(self, task, config):
        if config is False:  # handles 'template: no' form to turn off template on this task
            return
        # implements --template NAME
        if task.options.template:
            if not config or task.options.template not in config:
                task.abort('does not use `%s` template' % task.options.template, silent=True)

        config = self.prepare_config(config)

        # add global in except when disabled with no_global
        if 'no_global' in config:
            config.remove('no_global')
            if 'global' in config:
                config.remove('global')
        elif 'global' not in config:
            config.append('global')

        toplevel_templates = task.manager.config.get('templates', {})

        # apply templates
        for template in config:
            if template not in toplevel_templates:
                if template == 'global':
                    continue
                raise plugin.PluginError(
                    'Unable to find template %s for task %s' % (template, task.name), logger
                )
            if toplevel_templates[template] is None:
                logger.warning('Template `{}` is empty. Nothing to merge.', template)
                continue
            logger.debug('Merging template {} into task {}', template, task.name)

            # We make a copy here because we need to remove
            template_config = toplevel_templates[template]
            # When there are templates within templates we remove the template
            # key from the config and append it's items to our own
            if 'template' in template_config:
                nested_templates = self.prepare_config(template_config['template'])
                for nested_template in nested_templates:
                    if nested_template not in config:
                        config.append(nested_template)
                    else:
                        logger.warning('Templates contain each other in a loop.')
                # Replace template_config with a copy without the template key, to avoid merging errors
                template_config = dict(template_config)
                del template_config['template']

            # Merge
            try:
                task.merge_config(template_config)
            except MergeException as exc:
                raise plugin.PluginError(
                    'Failed to merge template %s to task %s. Error: %s'
                    % (template, task.name, exc.value)
                )

        logger.trace('templates: {}', config)


@event('plugin.register')
def register_plugin():
    plugin.register(PluginTemplate, 'template', builtin=True, api_ver=2)


@event('config.register')
def register_config():
    root_config_schema = {
        'type': 'object',
        'additionalProperties': plugin.plugin_schemas(interface='task'),
    }
    register_config_key('templates', root_config_schema)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '-T', '--template', metavar='NAME', help='execute tasks using given template'
    )



import logging
import os
import re
import sys
import time

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils import qualities
from flexget.utils.parsers.generic import ParseWarning, default_ignore_prefixes, name_to_re
from flexget.utils.tools import ReList

from .parser_common import MovieParseResult, SeriesParseResult

# rebulk (that underlies guessit) will use the 'regex' module rather than 're' if installed.
# For consistency, prevent that unless env variable is explicitly already enabling it.
os.environ.setdefault('REGEX_DISABLED', 'true')  # isort:skip
from guessit.api import GuessItApi, GuessitException  # isort:skip
from guessit.rules import rebulk_builder  # isort:skip
from rebulk import Rebulk  # isort:skip
from rebulk.pattern import RePattern  # isort:skip


logger = logger.bind(name='parser_guessit')

logging.getLogger('rebulk').setLevel(logging.WARNING)
logging.getLogger('guessit').setLevel(logging.WARNING)


def _id_regexps_function(input_string, context):
    ret = []
    for regexp in context.get('id_regexps'):
        for match in RePattern(regexp, children=True).matches(input_string, context):
            ret.append(match.span)
    return ret


_id_regexps = Rebulk().functional(
    _id_regexps_function, name='regexpId', disabled=lambda context: not context.get('id_regexps')
)


def rules_builder(config):
    rebulk = rebulk_builder(config)
    rebulk.rebulk(_id_regexps)
    return rebulk


guessit_api = GuessItApi()
guessit_api.configure(options={}, rules_builder=rules_builder, force=True)


def normalize_component(data):
    if data is None:
        return []
    if isinstance(data, list):
        return [d.lower().replace('-', '') for d in data]

    return [data.lower().replace('-', '')]


try:
    preferred_clock = time.process_time
except AttributeError:
    preferred_clock = time.clock


class ParserGuessit:
    SOURCE_MAP = {
        'Camera': 'cam',
        'HD Camera': 'cam',
        'HD Telesync': 'telesync',
        'Pay-per-view': 'ppv',
        'Digital TV': 'dvb',
        'Video on Demand': 'vod',
        'Analog HDTV': 'ahdtv',
        'Ultra HDTV': 'uhdtv',
        'HD Telecine': 'hdtc',
        'Web': 'web-dl',
    }

    @staticmethod
    def _guessit_options(options):
        settings = {
            'name_only': True,
            'allowed_languages': ['en', 'fr'],
            'allowed_countries': ['us', 'uk', 'gb'],
            'single_value': True,
        }
        options['episode_prefer_number'] = not options.get('identified_by') == 'ep'
        if options.get('allow_groups'):
            options['expected_group'] = options['allow_groups']
        if 'date_yearfirst' in options:
            options['date_year_first'] = options['date_yearfirst']
        if 'date_dayfirst' in options:
            options['date_day_first'] = options['date_dayfirst']
        else:
            # See https://github.com/guessit-io/guessit/issues/329
            # https://github.com/guessit-io/guessit/pull/333
            # They made changes that break backward compatibility, so we have to make do this hackery
            if options.get('date_year_first'):
                options['date_day_first'] = True
        settings.update(options)
        return settings

    @staticmethod
    def _proper_count(guessit_result):
        """Calculate a FlexGet style proper_count from a guessit result."""
        version = guessit_result.get('version')
        if version is None:
            version = 0
        elif version <= 0:
            version = -1
        else:
            version -= 1
        proper_count = guessit_result.get('proper_count', 0)
        fastsub = 'fast subtitled' in normalize_component(guessit_result.values_list.get('other'))
        return version + proper_count - (5 if fastsub else 0)

    def _source(self, guessit_result):
        other = normalize_component(guessit_result.values_list.get('other'))
        source = self.SOURCE_MAP.get(guessit_result.get('source'), guessit_result.get('source'))
        # special case
        if source == 'web-dl' and 'rip' in other:
            source = 'webrip'

        source = normalize_component(source)

        if 'preair' in other:
            source.append('preair')
        if 'screener' in other:
            if 'bluray' in source:
                source.append('bdscr')
            else:
                source.append('dvdscr')
        if 'region 5' in other or 'region c' in other:
            source.append('r5')

        return source

    def _quality(self, guessit_result):
        """Generate a FlexGet Quality from a guessit result."""
        resolution = normalize_component(guessit_result.values_list.get('screen_size'))
        other = normalize_component(guessit_result.values_list.get('other'))
        if not resolution and 'high resolution' in other:
            resolution.append('hr')

        source = self._source(guessit_result)

        codec = normalize_component(guessit_result.values_list.get('video_codec'))
        if '10bit' in normalize_component(guessit_result.values_list.get('color_depth')):
            codec.append('10bit')

        audio = normalize_component(guessit_result.values_list.get('audio_codec'))
        audio_profile = normalize_component(guessit_result.values_list.get('audio_profile'))
        audio_channels = normalize_component(guessit_result.values_list.get('audio_channels'))
        # unlike the other components, audio can be a bit iffy with multiple codecs, so we limit it to one
        if 'dts' in audio and any(hd in audio_profile for hd in ['hd', 'master audio']):
            audio = ['dtshd']
        elif '5.1' in audio_channels and 'dolby digital plus' in audio:
            audio = ['dd+5.1']
        elif '5.1' in audio_channels and 'dolby digital' in audio:
            audio = ['dd5.1']

        # Make sure everything are strings (guessit will return lists when there are multiples)
        flattened_qualities = []
        for component in (resolution, source, codec, audio):
            if isinstance(component, list):
                flattened_qualities.append(' '.join(component))
            elif isinstance(component, str):
                flattened_qualities.append(component)
            else:
                raise ParseWarning(
                    self,
                    'Guessit quality returned type {}: {}. Expected str or list.'.format(
                        type(component), component
                    ),
                )

        return qualities.Quality(' '.join(flattened_qualities))

    # movie_parser API
    def parse_movie(self, data, **kwargs):
        logger.debug('Parsing movie: `{}` [options: {}]', data, kwargs)
        start = preferred_clock()
        guessit_options = self._guessit_options(kwargs)
        guessit_options['type'] = 'movie'
        guess_result = guessit_api.guessit(data, options=guessit_options)
        # NOTE: Guessit expects str on PY3 and unicode on PY2 hence the use of future.utils.native
        parsed = MovieParseResult(
            data=data,
            name=guess_result.get('title'),
            year=guess_result.get('year'),
            proper_count=self._proper_count(guess_result),
            quality=self._quality(guess_result),
            release_group=guess_result.get('release_group'),
            valid=bool(
                guess_result.get('title')
            ),  # It's not valid if it didn't find a name, which sometimes happens
        )
        logger.debug('Parsing result: {} (in {} ms)', parsed, (preferred_clock() - start) * 1000)
        return parsed

    # series_parser API
    def parse_series(self, data, **kwargs):
        logger.debug('Parsing series: `{}` [options: {}]', data, kwargs)
        guessit_options = self._guessit_options(kwargs)
        valid = True
        if kwargs.get('name'):
            expected_titles = [kwargs['name']]
            if kwargs.get('alternate_names'):
                expected_titles.extend(kwargs['alternate_names'])
            # apostrophe support
            expected_titles = [
                title.replace('\'', '(?:\'|\\\'|\\\\\'|-|)?') for title in expected_titles
            ]
            guessit_options['expected_title'] = ['re:' + title for title in expected_titles]
        if kwargs.get('id_regexps'):
            guessit_options['id_regexps'] = kwargs.get('id_regexps')
        start = preferred_clock()
        # If no series name is provided, we don't tell guessit what kind of match we are looking for
        # This prevents guessit from determining that too general of matches are series
        parse_type = 'episode' if kwargs.get('name') else None
        if parse_type:
            guessit_options['type'] = parse_type

        # NOTE: Guessit expects str on PY3 and unicode on PY2 hence the use of future.utils.native
        try:
            guess_result = guessit_api.guessit(data, options=guessit_options)
        except GuessitException:
            logger.warning('Parsing {} with guessit failed. Most likely a unicode error.', data)
            return SeriesParseResult(data=data, valid=False)

        if guess_result.get('type') != 'episode':
            valid = False

        name = kwargs.get('name')
        country = guess_result.get('country')
        if not name:
            name = guess_result.get('title', '')
            if not name:
                valid = False
            elif country and hasattr(country, 'alpha2'):
                name += ' (%s)' % country.alpha2
        elif guess_result.matches['title']:
            # Make sure the name match is up to FlexGet standards
            # Check there is no unmatched cruft before the matched name
            title_start = guess_result.matches['title'][0].start
            title_end = guess_result.matches['title'][0].end
            if title_start != 0:
                try:
                    pre_title = max(
                        (
                            match[0].end
                            for match in guess_result.matches.values()
                            if match[0].end <= title_start
                        )
                    )
                except ValueError:
                    pre_title = 0
                for char in reversed(data[pre_title:title_start]):
                    if char.isalnum() or char.isdigit():
                        return SeriesParseResult(data=data, valid=False)
                    if char.isspace() or char in '._':
                        continue
                    else:
                        break
            # Check the name doesn't end mid-word (guessit might put the border before or after the space after title)
            if (
                data[title_end - 1].isalnum()
                and len(data) <= title_end
                or not self._is_valid_name(data, guessit_options=guessit_options)
            ):
                valid = False
            # If we are in exact mode, make sure there is nothing after the title
            if kwargs.get('strict_name'):
                post_title = sys.maxsize
                for match_type, matches in guess_result.matches.items():
                    if match_type in ['season', 'episode', 'date', 'regexpId']:
                        if matches[0].start < title_end:
                            continue
                        post_title = min(post_title, matches[0].start)
                        if matches[0].parent:
                            post_title = min(post_title, matches[0].parent.start)
                for char in data[title_end:post_title]:
                    if char.isalnum() or char.isdigit():
                        valid = False
        else:
            valid = False
        season = guess_result.get('season')
        episode = guess_result.get('episode')
        if episode is None and 'part' in guess_result:
            episode = guess_result['part']
        if isinstance(episode, list):
            # guessit >=2.1.4 returns a list for multi-packs, but we just want the first one and the number of eps
            episode = episode[0]
        date = guess_result.get('date')
        quality = self._quality(guess_result)
        proper_count = self._proper_count(guess_result)
        group = guess_result.get('release_group')
        # Validate group with from_group
        if not self._is_valid_groups(group, guessit_options.get('allow_groups', [])):
            valid = False
        # Validate country, TODO: LEGACY
        if country and name.endswith(')'):
            p_start = name.rfind('(')
            if p_start != -1:
                parenthetical = re.escape(name[p_start + 1 : -1])
                if parenthetical and parenthetical.lower() != str(country).lower():
                    valid = False
        # Check the full list of 'episode_details' for special,
        # since things like 'pilot' and 'unaired' can also show up there
        special = any(
            v.lower() == 'special' for v in guess_result.values_list.get('episode_details', [])
        )
        if 'episode' not in guess_result.values_list:
            episodes = len(guess_result.values_list.get('part', []))
        else:
            episodes = len(guess_result.values_list['episode'])
        if episodes > 3:
            valid = False
        identified_by = kwargs.get('identified_by', 'auto')
        identifier_type, identifier = None, None
        if identified_by in ['date', 'auto']:
            if date:
                identifier_type = 'date'
                identifier = date
        if not identifier_type and identified_by in ['ep', 'auto']:
            if episode is not None:
                if season is None and kwargs.get('allow_seasonless', True):
                    if 'part' in guess_result:
                        season = 1
                    else:
                        episode_raw = guess_result.matches['episode'][0].initiator.raw
                        if episode_raw and any(
                            c.isalpha() and c.lower() != 'v' for c in episode_raw
                        ):
                            season = 1
                if season is not None:
                    identifier_type = 'ep'
                    identifier = (season, episode)

        if not identifier_type and identified_by in ['id', 'auto']:
            if guess_result.matches['regexpId']:
                identifier_type = 'id'
                identifier = '-'.join(match.value for match in guess_result.matches['regexpId'])
        if not identifier_type and identified_by in ['sequence', 'auto']:
            if episode is not None:
                identifier_type = 'sequence'
                identifier = episode
        if (not identifier_type or guessit_options.get('prefer_specials')) and (
            special or guessit_options.get('assume_special')
        ):
            identifier_type = 'special'
            identifier = guess_result.get('episode_title', 'special')
        if not identifier_type:
            valid = False
        # TODO: Legacy - Complete == invalid
        if 'complete' in normalize_component(guess_result.get('other')):
            valid = False

        parsed = SeriesParseResult(
            data=data,
            name=name,
            episodes=episodes,
            identified_by=identified_by,
            id=identifier,
            id_type=identifier_type,
            quality=quality,
            proper_count=proper_count,
            special=special,
            group=group,
            valid=valid,
        )

        logger.debug('Parsing result: {} (in {} ms)', parsed, (preferred_clock() - start) * 1000)
        return parsed

    # TODO: The following functions are sort of legacy. No idea if they should be changed.
    def _is_valid_name(self, data, guessit_options):
        if not guessit_options.get('name'):
            return True
        # name end position
        name_end = 0

        # regexp name matching
        re_from_name = False
        name_regexps = ReList(guessit_options.get('name_regexps', []))
        if not name_regexps:
            # if we don't have name_regexps, generate one from the name
            name_regexps = ReList(
                name_to_re(name, default_ignore_prefixes, None)
                for name in [guessit_options['name']] + guessit_options.get('alternate_names', [])
            )
            # With auto regex generation, the first regex group captures the name
            re_from_name = True
        # try all specified regexps on this data
        for name_re in name_regexps:
            match = re.search(name_re, data)
            if match:
                match_end = match.end(1 if re_from_name else 0)
                # Always pick the longest matching regex
                if match_end > name_end:
                    name_end = match_end
                logger.debug('NAME SUCCESS: {} matched to {}', name_re.pattern, data)
        if not name_end:
            # leave this invalid
            logger.debug(
                'FAIL: name regexps {} do not match {}',
                [regexp.pattern for regexp in name_regexps],
                data,
            )
            return False
        return True

    def _is_valid_groups(self, group, allow_groups):
        if not allow_groups:
            return True
        if not group:
            return False
        normalized_allow_groups = [x.lower() for x in allow_groups]
        # TODO: special case for guessit with expected_group parameter
        if isinstance(group, list):
            return any(g.lower() in normalized_allow_groups for g in group)

        return group.lower() in normalized_allow_groups


@event('plugin.register')
def register_plugin():
    plugin.register(
        ParserGuessit, 'parser_guessit', interfaces=['movie_parser', 'series_parser'], api_ver=2
    )



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='redirect_url')


class UrlRewriteRedirect:
    """Rewrites urls which actually redirect somewhere else."""

    def __init__(self):
        self.processed = set()

    def on_task_start(self, task, config):
        self.processed = set()

    def on_task_urlrewrite(self, task, config):
        if not config:
            return
        for entry in task.accepted:
            if not any(entry['url'].startswith(adapter) for adapter in task.requests.adapters):
                continue
            elif entry['url'] in self.processed:
                continue
            auth = None
            if 'download_auth' in entry:
                auth = entry['download_auth']
                logger.debug(
                    'Custom auth enabled for {} url_redirect: {}',
                    entry['title'],
                    entry['download_auth'],
                )
            try:
                r = task.requests.head(entry['url'], auth=auth, allow_redirects=True)
            except Exception:
                pass
            else:
                if r.status_code < 400 and r.url != entry['url']:
                    entry['url'] = r.url
            # Make sure we don't try to rewrite this url again
            self.processed.add(entry['url'])


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteRedirect, 'redirect_url', api_ver=2)



# Hack, hide DataLossWarnings
# Based on html5lib code namespaceHTMLElements=False should do it, but nope ...
# Also it doesn't seem to be available in older version from html5lib, removing it
import warnings
from typing import IO, Union

from bs4 import BeautifulSoup
from html5lib.constants import DataLossWarning

warnings.simplefilter('ignore', DataLossWarning)


def get_soup(obj: Union[str, IO, bytes], parser: str = 'html5lib') -> BeautifulSoup:
    return BeautifulSoup(obj, parser)



import os

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.template import RenderError, get_template, render_from_task

PLUGIN_NAME = 'make_html'

logger = logger.bind(name=PLUGIN_NAME)


class OutputHtml:
    schema = {
        'type': 'object',
        'properties': {'template': {'type': 'string'}, 'file': {'type': 'string'}},
        'required': ['file'],
        'additionalProperties': False,
    }

    def on_task_output(self, task, config):
        # Use the default template if none is specified
        if not config.get('template'):
            config['template'] = 'html.template'

        filename = os.path.expanduser(config['template'])
        output = os.path.expanduser(config['file'])
        # Output to config directory if absolute path has not been specified
        if not os.path.isabs(output):
            output = os.path.join(task.manager.config_base, output)

        # create the template
        try:
            template = render_from_task(get_template(filename), task)
            logger.verbose('Writing output html to {}', output)
            with open(output, 'wb') as f:
                f.write(template.encode('utf-8'))
        except RenderError as e:
            logger.error('Error while rendering task {}, Error: {}', task, e)
            raise plugin.PluginError('There was an error rendering the specified template')


@event('plugin.register')
def register_plugin():
    plugin.register(OutputHtml, PLUGIN_NAME, api_ver=2)



from urllib.parse import urlparse

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='shortened')


class UrlRewriteShortened:
    """Shortened url rewriter."""

    def url_rewritable(self, task, entry):
        return urlparse(entry['url']).netloc in ['bit.ly', 't.co']

    def url_rewrite(self, task, entry):
        request = task.requests.head(entry['url'], allow_redirects=True)
        entry['url'] = request.url


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteShortened, 'shortened', interfaces=['urlrewriter'], api_ver=2)



import datetime

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.database import Session

logger = logger.bind(name='est_movies_bluray')


class EstimatesMoviesBluray:
    @plugin.priority(2)
    def estimate(self, entry):
        if 'movie_name' not in entry:
            return

        movie_name = entry['movie_name']
        movie_year = entry.get('movie_year')

        if movie_year is not None and movie_year > datetime.datetime.now().year:
            logger.debug('Skipping Blu-ray.com lookup since movie year is {}', movie_year)
            return

        logger.debug('Searching Blu-ray.com for release date of {} ({})', movie_name, movie_year)

        release_date = None
        try:
            with Session() as session:
                lookup = plugin.get('api_bluray', self).lookup
                movie = lookup(title=movie_name, year=movie_year, session=session)
                if movie:
                    release_date = movie.release_date
        except LookupError as e:
            logger.debug(e)
        if release_date:
            logger.debug('received release date: {}', release_date)
        return release_date


@event('plugin.register')
def register_plugin():
    plugin.register(
        EstimatesMoviesBluray, 'est_movies_bluray', interfaces=['estimate_release'], api_ver=2
    )



from flexget.components.managed_lists.lists.movie_list.api import ObjectsContainer as OC
from flexget.utils import json


class TestETAG:
    config = 'tasks: {}'

    def test_etag(self, api_client, schema_match):
        # Test ETag creation and usage

        # Create movie lists
        list_1 = {'name': 'list_1'}
        list_2 = {'name': 'list_2'}

        # Create lists
        rsp = api_client.json_post('/movie_list/', data=json.dumps(list_1))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        rsp = api_client.json_post('/movie_list/', data=json.dumps(list_2))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code

        # Get ETag
        rsp = api_client.get('/movie_list/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        etag = rsp.headers.get('etag')
        assert etag is not None

        # Test If-None-Match
        header = {'If-None-Match': etag}
        rsp = api_client.head('/movie_list/', headers=header)
        assert rsp.status_code == 304, 'Response code is %s' % rsp.status_code

        header = {'If-None-Match': etag}
        rsp = api_client.get('/movie_list/', headers=header)
        assert rsp.status_code == 304, 'Response code is %s' % rsp.status_code
        data = rsp.get_data(as_text=True)
        assert data is ''

        header = {'If-None-Match': '*'}
        rsp = api_client.head('/movie_list/', headers=header)
        assert rsp.status_code == 304, 'Response code is %s' % rsp.status_code

        # Test If-Match
        header = {'If-Match': 'not_etag'}
        rsp = api_client.head('/movie_list/', headers=header)
        assert rsp.status_code == 412, 'Response code is %s' % rsp.status_code

        # Change data
        list_3 = {'name': 'list_3'}
        rsp = api_client.json_post('/movie_list/', data=json.dumps(list_3))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code

        header = {'If-None-Match': etag}
        rsp = api_client.get('/movie_list/', headers=header)
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(OC.return_lists, data)
        assert not errors

        # Verify all 3 lists are received as payload
        assert len(data) == 3



import re
from argparse import ArgumentParser, ArgumentTypeError

from flexget import options
from flexget.event import event
from flexget.terminal import TerminalTable, TerminalTableError, console, table_parser
from flexget.utils.database import Session

from . import db


def do_cli(manager, options):
    """Handle regexp-list cli"""
    action_map = {
        'all': action_all,
        'list': action_list,
        'add': action_add,
        'del': action_del,
        'purge': action_purge,
    }

    action_map[options.regexp_action](options)


def action_all(options):
    """ Show all regexp lists """
    lists = db.get_regexp_lists()
    header = ['#', 'List Name']
    table_data = [header]
    for regexp_list in lists:
        table_data.append([regexp_list.id, regexp_list.name])
    table = TerminalTable(options.table_type, table_data)
    try:
        console(table.output)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))


def action_list(options):
    """List regexp list"""
    with Session() as session:
        regexp_list = db.get_list_by_exact_name(options.list_name)
        if not regexp_list:
            console('Could not find regexp list with name {}'.format(options.list_name))
            return
        header = ['Regexp']
        table_data = [header]
        regexps = db.get_regexps_by_list_id(
            regexp_list.id, order_by='added', descending=True, session=session
        )
        for regexp in regexps:
            regexp_row = [regexp.regexp or '']
            table_data.append(regexp_row)
    try:
        table = TerminalTable(options.table_type, table_data)
        console(table.output)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))


def action_add(options):
    with Session() as session:
        regexp_list = db.get_list_by_exact_name(options.list_name)
        if not regexp_list:
            console('Could not find regexp list with name {}, creating'.format(options.list_name))
            regexp_list = db.create_list(options.list_name, session=session)

        regexp = db.get_regexp(list_id=regexp_list.id, regexp=options.regexp, session=session)
        if not regexp:
            console("Adding regexp {} to list {}".format(options.regexp, regexp_list.name))
            db.add_to_list_by_name(regexp_list.name, options.regexp, session=session)
            console(
                'Successfully added regexp {} to regexp list {} '.format(
                    options.regexp, regexp_list.name
                )
            )
        else:
            console("Regexp {} already exists in list {}".format(options.regexp, regexp_list.name))


def action_del(options):
    with Session() as session:
        regexp_list = db.get_list_by_exact_name(options.list_name)
        if not regexp_list:
            console('Could not find regexp list with name {}'.format(options.list_name))
            return
        regexp = db.get_regexp(list_id=regexp_list.id, regexp=options.regexp, session=session)
        if regexp:
            console('Removing regexp {} from list {}'.format(options.regexp, options.list_name))
            session.delete(regexp)
        else:
            console(
                'Could not find regexp {} in list {}'.format(
                    options.movie_title, options.list_name
                )
            )
            return


def action_purge(options):
    with Session() as session:
        regexp_list = db.get_list_by_exact_name(options.list_name)
        if not regexp_list:
            console('Could not find regexp list with name {}'.format(options.list_name))
            return
        console('Deleting list %s' % options.list_name)
        session.delete(regexp_list)


def regexp_type(regexp):
    try:
        re.compile(regexp)
        return regexp
    except re.error as e:
        raise ArgumentTypeError(e)


@event('options.register')
def register_parser_arguments():
    # Common option to be used in multiple subparsers
    regexp_parser = ArgumentParser(add_help=False)
    regexp_parser.add_argument('regexp', type=regexp_type, help="The regexp")

    list_name_parser = ArgumentParser(add_help=False)
    list_name_parser.add_argument(
        'list_name', nargs='?', help='Name of regexp list to operate on', default='regexps'
    )
    # Register subcommand
    parser = options.register_command('regexp-list', do_cli, help='View and manage regexp lists')
    # Set up our subparsers
    subparsers = parser.add_subparsers(title='actions', metavar='<action>', dest='regexp_action')
    subparsers.add_parser('all', parents=[table_parser], help='Shows all existing regexp lists')
    subparsers.add_parser(
        'list', parents=[list_name_parser, table_parser], help='List regexp from a list'
    )
    subparsers.add_parser(
        'add', parents=[list_name_parser, regexp_parser], help='Add a regexp to a list'
    )
    subparsers.add_parser(
        'del', parents=[list_name_parser, regexp_parser], help='Remove a regexp from a list'
    )
    subparsers.add_parser(
        'purge', parents=[list_name_parser], help='Removes an entire list. Use with caution!'
    )



import datetime
import os
import re
from collections import defaultdict
from typing import Any, Callable, Dict, List, Match, Optional, Pattern, Union
from urllib.parse import parse_qsl, urlparse

import jsonschema
from jsonschema import ValidationError
from jsonschema.compat import int_types, str_types
from loguru import logger

from flexget.event import fire_event
from flexget.utils import qualities, template
from flexget.utils.template import get_template
from flexget.utils.tools import parse_episode_identifier, parse_timedelta

logger = logger.bind(name='config_schema')

CURRENT_SCHEMA_VERSION = 'http://json-schema.org/draft-04/schema#'
# Type hint for json schemas. (If we upgrade to a newer json schema version, the type might allow more than dicts.)
JsonSchema = Dict[str, Any]
schema_paths: Dict[str, Union[JsonSchema, Callable[..., JsonSchema]]] = {}


class ConfigValidationError(ValidationError):
    json_pointer: str


class ConfigError(ValueError):
    errors: List[ConfigValidationError]


# TODO: Rethink how config key and schema registration work
def register_schema(path: str, schema: Union[JsonSchema, Callable[..., JsonSchema]]):
    """
    Register `schema` to be available at `path` for $refs

    :param path: Path to make schema available
    :param schema: The schema, or function which returns the schema
    """
    schema_paths[path] = schema


# Validator that handles root structure of config.
_root_config_schema: Optional[JsonSchema] = None


def register_config_key(key: str, schema: JsonSchema, required: bool = False):
    """Registers a valid root level key for the config.

    :param string key:
      Name of the root level key being registered.
    :param dict schema:
      Schema for the key.
    :param bool required:
      Specify whether this is a mandatory key.
    """
    root_schema = get_schema()
    root_schema['properties'][key] = schema
    if required:
        root_schema.setdefault('required', []).append(key)
    register_schema('/schema/config/%s' % key, schema)


def get_schema() -> JsonSchema:
    global _root_config_schema
    if _root_config_schema is None:
        _root_config_schema = {
            'type': 'object',
            'properties': {},
            'additionalProperties': False,
            '$schema': CURRENT_SCHEMA_VERSION,
        }
        fire_event('config.register')
        # TODO: Is /schema/config this the best place for this?
        register_schema('/schema/config', _root_config_schema)
    return _root_config_schema


def one_or_more(schema: JsonSchema, unique_items: bool = False) -> JsonSchema:
    """
    Helper function to construct a schema that validates items matching `schema` or an array
    containing items matching `schema`.

    """

    schema.setdefault('title', 'single value')
    default = schema.pop('default', None)
    result = {
        'oneOf': [
            {
                'title': 'multiple values',
                'type': 'array',
                'items': schema,
                'minItems': 1,
                'uniqueItems': unique_items,
            },
            schema,
        ]
    }
    if default:
        result['default'] = default
    return result


def resolve_ref(uri: str) -> JsonSchema:
    """
    Finds and returns a schema pointed to by `uri` that has been registered in the register_schema function.
    """
    parsed = urlparse(uri)
    if parsed.path in schema_paths:
        schema = schema_paths[parsed.path]
        if callable(schema):
            schema = schema(**dict(parse_qsl(parsed.query)))
        schema = {'$schema': CURRENT_SCHEMA_VERSION, **schema}
        return schema
    raise jsonschema.RefResolutionError("%s could not be resolved" % uri)


def process_config(
    config: Any, schema: Optional[JsonSchema] = None, set_defaults: bool = True
) -> List[ConfigValidationError]:
    """
    Validates the config, and sets defaults within it if `set_defaults` is set.
    If schema is not given, uses the root config schema.

    :returns: A list with :class:`jsonschema.ValidationError`s if any

    """
    if schema is None:
        schema = get_schema()
    resolver = RefResolver.from_schema(schema)
    validator = SchemaValidator(schema, resolver=resolver, format_checker=format_checker)
    if set_defaults:
        validator.VALIDATORS['properties'] = validate_properties_w_defaults
    try:
        errors: List[ValidationError] = list(validator.iter_errors(config))
    finally:
        validator.VALIDATORS['properties'] = jsonschema.Draft4Validator.VALIDATORS['properties']
    # Customize the error messages
    for e in errors:
        set_error_message(e)
        e.json_pointer = '/' + '/'.join(map(str, e.path))
    return errors


def parse_time(time_string: str) -> datetime.time:
    """Parse a time string from the config into a :class:`datetime.time` object."""
    formats = ['%I:%M %p', '%H:%M', '%H:%M:%S']
    for f in formats:
        try:
            return datetime.datetime.strptime(time_string, f).time()
        except ValueError:
            continue
    raise ValueError('invalid time `%s`' % time_string)


def parse_interval(interval_string: str) -> datetime.timedelta:
    """Takes an interval string from the config and turns it into a :class:`datetime.timedelta` object."""
    regexp = r'^\d+ (second|minute|hour|day|week)s?$'
    if not re.match(regexp, interval_string):
        raise ValueError("should be in format 'x (seconds|minutes|hours|days|weeks)'")
    return parse_timedelta(interval_string)


def parse_percent(percent_input: str) -> float:
    """Takes a percent string from the config and turns it into a float."""
    percent_input = percent_input.rstrip('%')
    try:
        return float(percent_input)
    except ValueError:
        raise ValueError("should be in format '0-x%'")


def parse_size(size_input: str) -> int:
    """Takes a size string from the config and turns it into int(bytes)."""
    prefixes = [None, 'K', 'M', 'G', 'T', 'P']
    try:
        # Bytes
        return int(size_input)
    except ValueError:
        size_input = size_input.upper().rstrip('IB')
        value, unit = float(size_input[:-1]), size_input[-1:]
        if unit not in prefixes:
            raise ValueError("should be in format '0-x (KiB, MiB, GiB, TiB, PiB)'")
        return int(1024 ** prefixes.index(unit) * value)


# Public API end here, the rest should not be used outside this module


class RefResolver(jsonschema.RefResolver):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('handlers', {'': resolve_ref})
        super().__init__(*args, **kwargs)


format_checker = jsonschema.FormatChecker(('email',))


@format_checker.checks('quality', raises=ValueError)
def is_quality(instance):
    if not isinstance(instance, str_types):
        return True
    return qualities.get(instance)


@format_checker.checks('quality_requirements', raises=ValueError)
def is_quality_req(instance):
    if not isinstance(instance, str_types):
        return True
    return qualities.Requirements(instance)


@format_checker.checks('time', raises=ValueError)
def is_time(time_string) -> bool:
    if not isinstance(time_string, str_types):
        return True
    return parse_time(time_string) is not None


@format_checker.checks('interval', raises=ValueError)
def is_interval(interval_string) -> bool:
    if not isinstance(interval_string, str_types):
        return True
    return parse_interval(interval_string) is not None


@format_checker.checks('size', raises=ValueError)
def is_size(size_string) -> bool:
    if not isinstance(size_string, (str_types, int_types)):
        return True
    return parse_size(size_string) is not None


@format_checker.checks('percent', raises=ValueError)
def is_percent(percent_string) -> bool:
    if not isinstance(percent_string, str_types):
        return True
    return parse_percent(percent_string) is not None


@format_checker.checks('regex', raises=ValueError)
def is_regex(instance) -> Union[bool, Pattern]:
    if not isinstance(instance, str_types):
        return True
    try:
        return re.compile(instance)
    except re.error as e:
        raise ValueError('Error parsing regex: %s' % e)


@format_checker.checks('file', raises=ValueError)
def is_file(instance) -> bool:
    if not isinstance(instance, str_types):
        return True
    if os.path.isfile(os.path.expanduser(instance)):
        return True
    raise ValueError('`%s` does not exist' % instance)


@format_checker.checks('path', raises=ValueError)
def is_path(instance) -> bool:
    if not isinstance(instance, str_types):
        return True
    # Only validate the part of the path before the first identifier to be replaced
    pat = re.compile(r'{[{%].*[}%]}')
    result = pat.search(instance)
    if result:
        instance = os.path.dirname(instance[0 : result.start()])
    if os.path.isdir(os.path.expanduser(instance)):
        return True
    raise ValueError('`%s` does not exist' % instance)


# TODO: jsonschema has a format checker for uri if rfc3987 is installed, perhaps we should use that
@format_checker.checks('url')
def is_url(instance) -> Union[None, bool, Match]:
    if not isinstance(instance, str_types):
        return True
    regexp = (
        '('
        + '|'.join(['ftp', 'http', 'https', 'file', 'udp', 'socks5h?'])
        + r'):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%@!\-\/]))?'
    )
    return re.match(regexp, instance)


@format_checker.checks('episode_identifier', raises=ValueError)
def is_episode_identifier(instance) -> bool:
    if not isinstance(instance, (str_types, int)):
        return True
    return parse_episode_identifier(instance) is not None


@format_checker.checks('episode_or_season_id', raises=ValueError)
def is_episode_or_season_id(instance):
    if not isinstance(instance, (str_types, int)):
        return True
    return parse_episode_identifier(instance, identify_season=True) is not None


@format_checker.checks('file_template', raises=ValueError)
def is_valid_template(instance) -> bool:
    if not isinstance(instance, str_types):
        return True
    return get_template(instance) is not None


def set_error_message(error: jsonschema.ValidationError) -> None:
    """
    Create user facing error message from a :class:`jsonschema.ValidationError` `error`

    """
    # First, replace default error messages with our custom ones
    if error.validator == 'type':
        if isinstance(error.validator_value, str):
            valid_types_list = [error.validator_value]
        else:
            valid_types_list = list(error.validator_value)
        # Replace some types with more pythony ones
        replace = {'object': 'dict', 'array': 'list'}
        valid_types_list = [replace.get(t, t) for t in valid_types_list]
        # Make valid_types into an english list, with commas and 'or'
        valid_types = ', '.join(valid_types_list[:-2] + ['']) + ' or '.join(valid_types_list[-2:])
        if isinstance(error.instance, dict):
            error.message = f'Got a dict, expected: {valid_types}'
        if isinstance(error.instance, list):
            error.message = f'Got a list, expected: {valid_types}'
        error.message = f'Got `{error.instance}`, expected: {valid_types}'
    elif error.validator == 'format':
        if error.cause:
            error.message = str(error.cause)
    elif error.validator == 'enum':
        error.message = 'Must be one of the following: %s' % ', '.join(
            map(str, error.validator_value)
        )
    elif error.validator == 'additionalProperties':
        if error.validator_value is False:
            extras = set(
                jsonschema._utils.find_additional_properties(error.instance, error.schema)
            )
            if len(extras) == 1:
                error.message = 'The key `%s` is not valid here.' % extras.pop()
            else:
                error.message = 'The keys %s are not valid here.' % ', '.join(
                    '`%s`' % e for e in extras
                )
    else:
        # Remove u'' string representation from jsonschema error messages
        error.message = re.sub('u\'(.*?)\'', '`\\1`', error.message)

    # Then update with any custom error message supplied from the schema
    custom_error = error.schema.get('error_%s' % error.validator, error.schema.get('error'))
    if custom_error:
        error.message = template.render(custom_error, error.__dict__)


def select_child_errors(validator, errors):
    """
    Looks through subschema errors, if any subschema is determined to be the intended one,
    (based on 'type' keyword errors,) errors from its branch will be released instead of the parent error.
    """
    for error in errors:
        if not error.context:
            yield error
            continue
        # Split the suberrors up by which subschema they are from
        subschema_errors = defaultdict(list)
        for sube in error.context:
            subschema_errors[sube.schema_path[0]].append(sube)
        # Find the subschemas that did not have a 'type' error validating the instance at this path
        no_type_errors = dict(subschema_errors)
        valid_types = set()
        for i, errors in subschema_errors.items():
            for e in errors:
                if e.validator == 'type' and not e.path:
                    # Remove from the no_type_errors dict
                    no_type_errors.pop(i, None)
                    # Add the valid types to the list of all valid types
                    if validator.is_type(e.validator_value, 'string'):
                        valid_types.add(e.validator_value)
                    else:
                        valid_types.update(e.validator_value)
        if not no_type_errors:
            # If all of the branches had a 'type' error, create our own virtual type error with all possible types
            for e in validator.descend(error.instance, {'type': valid_types}):
                yield e
        elif len(no_type_errors) == 1:
            # If one of the possible schemas did not have a 'type' error, assume that is the intended one and issue
            # all errors from that subschema
            for e in list(no_type_errors.values())[0]:
                e.schema_path.extendleft(reversed(error.schema_path))
                e.path.extendleft(reversed(error.path))
                yield e
        else:
            yield error


def validate_properties_w_defaults(validator, properties, instance, schema):
    if not validator.is_type(instance, 'object'):
        return
    for key, subschema in properties.items():
        if 'default' in subschema:
            instance.setdefault(key, subschema['default'])
    for error in jsonschema.Draft4Validator.VALIDATORS["properties"](
        validator, properties, instance, schema
    ):
        yield error


def validate_anyOf(validator, anyOf, instance, schema):
    errors = jsonschema.Draft4Validator.VALIDATORS["anyOf"](validator, anyOf, instance, schema)
    for e in select_child_errors(validator, errors):
        yield e


def validate_oneOf(validator, oneOf, instance, schema):
    errors = jsonschema.Draft4Validator.VALIDATORS["oneOf"](validator, oneOf, instance, schema)
    for e in select_child_errors(validator, errors):
        yield e


def validate_deprecated(validator, message, instance, schema):
    """Not really a validator, just warns if deprecated section of config is being used."""
    logger.warning(message)


validators = {'anyOf': validate_anyOf, 'oneOf': validate_oneOf, 'deprecated': validate_deprecated}

SchemaValidator = jsonschema.validators.extend(jsonschema.Draft4Validator, validators)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='inputs')


class PluginInputs:
    """
    Allows the same input plugin to be configured multiple times in a task.

    Example::

      inputs:
        - rss: http://feeda.com
        - rss: http://feedb.com
    """

    schema = {
        'type': 'array',
        'items': {
            'allOf': [
                {'$ref': '/schema/plugins?phase=input'},
                {
                    'maxProperties': 1,
                    'error_maxProperties': 'Plugin options within inputs plugin must be indented 2 more spaces than '
                    'the first letter of the plugin name.',
                    'minProperties': 1,
                },
            ]
        },
    }

    def on_task_input(self, task, config):
        entry_titles = set()
        entry_urls = set()
        for item in config:
            for input_name, input_config in item.items():
                input = plugin.get_plugin_by_name(input_name)
                method = input.phase_handlers['input']
                try:
                    result = method(task, input_config)
                except plugin.PluginError as e:
                    logger.warning('Error during input plugin {}: {}', input_name, e)
                    continue
                if not result:
                    msg = 'Input %s did not return anything' % input_name
                    if getattr(task, 'no_entries_ok', False):
                        logger.verbose(msg)
                    else:
                        logger.warning(msg)
                    continue
                for entry in result:
                    if entry['title'] in entry_titles:
                        logger.debug('Title `{}` already in entry list, skipping.', entry['title'])
                        continue
                    urls = ([entry['url']] if entry.get('url') else []) + entry.get('urls', [])
                    if any(url in entry_urls for url in urls):
                        logger.debug(
                            'URL for `{}` already in entry list, skipping.', entry['title']
                        )
                        continue
                    yield entry
                    entry_titles.add(entry['title'])
                    entry_urls.update(urls)


@event('plugin.register')
def register_plugin():
    plugin.register(PluginInputs, 'inputs', api_ver=2)



from loguru import logger

from flexget import options, plugin
from flexget.event import event
from flexget.terminal import console

try:
    from guppy import hpy
except ImportError:
    # this will leave the plugin unloaded
    raise plugin.DependencyError(issued_by='memusage', missing='ext lib `guppy`', silent=True)

logger = logger.bind(name='mem_usage')

"""
http://blog.mfabrik.com/2008/03/07/debugging-django-memory-leak-with-trackrefs-and-guppy/

# Print memory statistics
def update():
    print heapy.heap()

# Print relative memory consumption since last sycle
def update():
    print heapy.heap()
    heapy.setref()

# Print relative memory consumption w/heap traversing
def update()
    print heapy.heap().get_rp(40)
    heapy.setref()
"""

heapy = None


@event('manager.startup')
def on_manager_startup(manager):
    if not manager.options.mem_usage:
        return
    global heapy
    heapy = hpy()


@event('manager.shutdown')
def on_manager_shutdown(manager):
    if not manager.options.mem_usage:
        return

    import resource

    console(
        'Resource Module memory usage: %s (kb)'
        % resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    )
    global heapy
    console('Heapy module calculating memory usage:')
    console(heapy.heap())
    console('-' * 79)
    console('Heapy module calculating report (this may take a while):')
    console(heapy.heap().get_rp(40))
    heapy = None


@event('options.register')
def register_parser_arguments():
    options.get_parser().add_argument(
        '--mem-usage',
        action='store_true',
        dest='mem_usage',
        default=False,
        help='display memory usage debug information',
    )



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.log import log_once

logger = logger.bind(name='thetvdb')


class FilterTvdb:
    """
    This plugin allows filtering based on thetvdb series rating,
    episode rating, status, genres, runtime, content-rating,
    languages, directors, writers, network, guest stars, episode
    rating, and actors

    Configuration:

    Note: All parameters are optional. Some are mutually exclusive.

    min_series_rating: <num>
    min_episode_rating: <num>
    min_episode_air_year: <num>
    max_episode_air_year: <num>
    min_episode_runtime: <num>
    max_episode_runtime: <num>

    # reject if genre contains any of these
    reject_content_rating:
        - TV-MA
    # accept only this content rating
    accept_content_rating:
        - TV-PG

    # accept only these networks
    accept_network:
        - NBC
    # reject if this network
    reject_network:
        - ABC

    # reject if genre contains any of these
    reject_genres:
        - drama
        - romance

    # reject if status contains any of these
    reject_status:
        - Ended

    # reject if language contain any of these
    reject_languages:
        - fr
    # accept only this language
    accept_languages:
        - en

    # Actors below take into account series actors, and guest stars
    # accept episode with any of these actors
    accept_actors:
        - Heidi Klum
        - Bruce Willis
    # reject episode if it has any of these actors
    reject_actors:
        - Cher
        - Tamala Jones

    # accept all episodes by these writers
    accept_writers:
        - Andrew W. Marlowe
    # reject episodes by these writers
    reject_writers:
        - Barry Schindel

    # accept all episodes by these directors
    accept_directors:
        - Rob Bowman
    # reject movies by these directors
    reject_directors:
        - John Terlesky
    """

    schema = {
        'type': 'object',
        'properties': {
            'min_series_rating': {'type': 'number'},
            'min_episode_rating': {'type': 'number'},
            'min_episode_air_year': {'type': 'integer'},
            'max_episode_air_year': {'type': 'integer'},
            'min_episode_runtime': {'type': 'number'},
            'max_episode_runtime': {'type': 'number'},
            'reject_content_rating': {'type': 'array', 'items': {'type': 'string'}},
            'accept_content_rating': {'type': 'array', 'items': {'type': 'string'}},
            'accept_network': {'type': 'array', 'items': {'type': 'string'}},
            'reject_network': {'type': 'array', 'items': {'type': 'string'}},
            'reject_genres': {'type': 'array', 'items': {'type': 'string'}},
            'reject_status': {'type': 'array', 'items': {'type': 'string'}},
            'accept_actors': {'type': 'array', 'items': {'type': 'string'}},
            'reject_actors': {'type': 'array', 'items': {'type': 'string'}},
            'accept_directors': {'type': 'array', 'items': {'type': 'string'}},
            'reject_directors': {'type': 'array', 'items': {'type': 'string'}},
        },
        'additionalProperties': False,
    }

    def is_in_set(self, config, configkey, entryitem):
        """
        this takes the config object, config key (to a list), and entry
        item so it can return True if the object matches,
        (be that a subset of the list, or if the entry item is contained
        within the config object list) or false if it does not.
        """
        # will want to port this over to filter_imdb as well, for code
        # clarity in that module.
        if configkey in config:
            configlist = config[configkey]
            if isinstance(entryitem, list):
                for item in entryitem:
                    if item in configlist:
                        return True
            else:
                if entryitem in configlist:
                    return True
        return False

    @plugin.priority(126)
    def on_task_filter(self, task, config):

        lookup = plugin.get('thetvdb_lookup', self).lookup

        for entry in task.entries:
            force_accept = False

            try:
                lookup(task, entry)
            except plugin.PluginError as e:
                logger.error('Skipping {} because of an error: {}', entry['title'], e.value)
                continue

            # Check defined conditions
            reasons = []
            if 'min_series_rating' in config:
                if entry['tvdb_rating'] < config['min_series_rating']:
                    reasons.append(
                        'series_rating (%s < %s)'
                        % (entry['tvdb_rating'], config['min_series_rating'])
                    )
            if 'min_episode_rating' in config:
                if entry['tvdb_ep_rating'] < config['min_episode_rating']:
                    reasons.append(
                        'tvdb_ep_rating (%s < %s)'
                        % (entry['tvdb_ep_rating'], config['min_episode_rating'])
                    )
            if 'min_episode_air_year' in config:
                if entry['tvdb_ep_air_date'].strftime("%Y") < config['min_episode_air_year']:
                    reasons.append(
                        'tvdb_ep_air_date (%s < %s)'
                        % (
                            entry['tvdb_ep_air_date'].strftime("%Y"),
                            config['min_episode_air_year'],
                        )
                    )
            if 'max_episode_air_year' in config:
                if entry['tvdb_ep_air_date'].strftime("%Y") > config['max_episode_air_year']:
                    reasons.append(
                        'tvdb_ep_air_date (%s < %s)'
                        % (
                            entry['tvdb_ep_air_date'].strftime("%Y"),
                            config['max_episode_air_year'],
                        )
                    )

            if self.is_in_set(config, 'reject_content_rating', entry['tvdb_content_rating']):
                reasons.append('reject_content_rating')

            if not self.is_in_set(config, 'accept_content_rating', entry['tvdb_content_rating']):
                reasons.append('accept_content_rating')

            if self.is_in_set(config, 'reject_network', entry['tvdb_network']):
                reasons.append('reject_network')

            if not self.is_in_set(config, 'accept_network', entry['tvdb_network']):
                reasons.append('accept_network')

            if self.is_in_set(config, 'reject_genres', entry['tvdb_genres']):
                reasons.append('reject_genres')

            if self.is_in_set(config, 'reject_status', entry['tvdb_status']):
                reasons.append('reject_status')

            # Accept if actors contains an accepted actor, but don't reject otherwise
            if self.is_in_set(
                config, 'accept_actors', entry['tvdb_actors'] + entry['tvdb_ep_guest_stars']
            ):
                force_accept = True

            if self.is_in_set(
                config, 'reject_actors', entry['tvdb_actors'] + entry['tvdb_ep_guest_stars']
            ):
                reasons.append('reject_genres')

            # Accept if director is an accepted director, but don't reject otherwise
            if self.is_in_set(config, 'accept_directors', entry['tvdb_ep_director']):
                force_accept = True

            if self.is_in_set(config, 'reject_directors', entry['tvdb_ep_director']):
                reasons.append('reject_directors')

            if reasons and not force_accept:
                msg = 'Skipping %s because of rule(s) %s' % (
                    entry.get('series_name_thetvdb', None) or entry['title'],
                    ', '.join(reasons),
                )
                if task.options.debug:
                    logger.debug(msg)
                else:
                    log_once(msg, logger)
            else:
                logger.debug('Accepting {}', entry)
                entry.accept()


@event('plugin.register')
def register_plugin():
    plugin.register(FilterTvdb, 'thetvdb', api_ver=2)



from argparse import ArgumentParser

from colorclass.toggles import disable_all_colors

from flexget import options
from flexget.event import event
from flexget.terminal import TerminalTable, TerminalTableError, colorize, console, table_parser

try:
    from irc_bot.simple_irc_bot import IRCChannelStatus, SimpleIRCBot
except ImportError:
    SimpleIRCBot = None
    IRCChannelStatus = None


def do_cli(manager, options):
    """Handle irc cli"""

    if SimpleIRCBot is None:
        console('irc_bot is not installed. install using `pip install irc_bot`.')
        return

    if hasattr(options, 'table_type') and options.table_type == 'porcelain':
        disable_all_colors()

    action_map = {'status': action_status, 'restart': action_restart, 'stop': action_stop}

    # NOTE: Direct importing of other plugins is discouraged
    from flexget.components.irc.irc import irc_manager

    if irc_manager is None:
        console('IRC daemon does not appear to be running.')
        return

    action_map[options.irc_action](options, irc_manager)


def action_status(options, irc_manager):
    connection = options.irc_connection
    try:
        status = irc_manager.status(connection)
    except ValueError as e:
        console('ERROR: %s' % e.args[0])
        return

    header = ['Name', 'Alive', 'Channels', 'Server']
    table_data = [header]

    for connection in status:
        for name, info in connection.items():
            alive = colorize('green', 'Yes') if info['alive'] else colorize('red', 'No')
            channels = []
            for channel in info['channels']:
                for channel_name, channel_status in channel.items():
                    channels.append(channel_name)
                    if channel_status == IRCChannelStatus.CONNECTED:
                        channels[-1] = colorize('green', '* ' + channels[-1])
            table_data.append(
                [name, alive, ', '.join(channels), '%s:%s' % (info['server'], info['port'])]
            )
    try:
        table = TerminalTable(options.table_type, table_data)
        console(table.output)
        console(colorize('green', ' * Connected channel'))
    except TerminalTableError as e:
        console('ERROR: %s' % e)


def action_restart(options, irc_manager):
    connection = options.irc_connection
    try:
        console('Restarting irc connection %s. It may take a short while.' % connection)
        irc_manager.restart_connections(connection)
        console(
            'Successfully restarted {0}. Use `flexget irc status {0}` to check its status.'.format(
                connection or 'all'
            )
        )
    except KeyError:
        console('ERROR: %s is not a valid irc connection' % connection)


def action_stop(options, irc_manager):
    connection = options.irc_connection
    try:
        console('Stopping irc connection %s. It may take a short while.' % connection)
        irc_manager.stop_connections(wait=False, name=connection)
        console(
            'Successfully stopped {0}. Use `flexget irc status {0}` to check its status.'.format(
                connection or 'all'
            )
        )
    except KeyError:
        console('ERROR: %s is not a valid irc connection' % connection)


@event('options.register')
def register_parser_arguments():
    # Common option to be used in multiple subparsers
    irc_parser = ArgumentParser(add_help=False)
    irc_parser.add_argument('irc_connection', nargs='?', help="Title of the irc connection")

    # Register subcommand
    parser = options.register_command('irc', do_cli, help='View and manage irc connections')
    # Set up our subparsers
    subparsers = parser.add_subparsers(title='actions', metavar='<action>', dest='irc_action')
    subparsers.add_parser(
        'status',
        parents=[irc_parser, table_parser],
        help='Shows status for specific irc connection',
    )
    subparsers.add_parser('restart', parents=[irc_parser], help='Restart an irc connection')
    subparsers.add_parser('stop', parents=[irc_parser], help='Stops an irc connection')



import hashlib
import logging
import os
import struct

import pytz
import tzlocal
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from loguru import logger

from flexget.config_schema import format_checker, register_config_key, register_schema
from flexget.event import event
from flexget.manager import manager
from flexget.utils import json

logger = logger.bind(name='scheduler')


# Add a format checker for more detailed errors on cron type schedules
@format_checker.checks('cron_schedule', raises=ValueError)
def is_cron_schedule(instance):
    if not isinstance(instance, dict):
        return True
    try:
        return CronTrigger(**instance)
    except TypeError:
        # A more specific error message about which key will also be shown by properties schema keyword
        raise ValueError('Invalid key for schedule.')


DEFAULT_SCHEDULES = [{'tasks': ['*'], 'interval': {'hours': 1}}]

UNITS = ['minutes', 'hours', 'days', 'weeks']
interval_schema = {
    'type': 'object',
    'title': 'Simple Interval',
    'properties': {
        'minutes': {'type': 'number'},
        'hours': {'type': 'number'},
        'days': {'type': 'number'},
        'weeks': {'type': 'number'},
        'jitter': {'type': 'integer'},
    },
    'anyOf': [{'required': [unit]} for unit in UNITS],
    'error_anyOf': 'Interval must be specified as one or more of %s' % ', '.join(UNITS),
    'additionalProperties': False,
}

cron_schema = {
    'type': 'object',
    'title': 'Advanced Cron Interval',
    'properties': {
        'year': {'type': ['integer', 'string']},
        'month': {'type': ['integer', 'string']},
        'day': {'type': ['integer', 'string']},
        'week': {'type': ['integer', 'string']},
        'day_of_week': {'type': ['integer', 'string']},
        'hour': {'type': ['integer', 'string']},
        'minute': {'type': ['integer', 'string']},
        'jitter': {'type': 'integer'},
    },
    'additionalProperties': False,
}

schedule_schema = {
    'type': 'object',
    'title': 'Schedule',
    'description': 'A schedule which runs specified tasks periodically.',
    'properties': {
        'tasks': {'type': ['array', 'string'], 'items': {'type': 'string'}},
        'interval': interval_schema,
        'schedule': cron_schema,
    },
    'required': ['tasks'],
    'minProperties': 2,
    'maxProperties': 2,
    'error_minProperties': 'Either `cron` or `interval` must be defined.',
    'error_maxProperties': 'Either `cron` or `interval` must be defined.',
    'additionalProperties': False,
}

main_schema = {
    'oneOf': [
        {'type': 'array', 'title': 'Enable', 'items': schedule_schema},
        {'type': 'boolean', 'title': 'Disable', 'description': 'Disable task schedules'},
    ]
}

scheduler = None
scheduler_job_map = {}


def job_id(conf):
    """Create a unique id for a schedule item in config."""
    return hashlib.sha1(json.dumps(conf, sort_keys=True).encode('utf-8')).hexdigest()


def run_job(tasks):
    """Add the execution to the queue and waits until it is finished"""
    logger.debug('executing tasks: {}', tasks)
    finished_events = manager.execute(
        options={'tasks': tasks, 'cron': True, 'allow_manual': False}, priority=5
    )
    for _, task_name, event_ in finished_events:
        logger.debug('task finished executing: {}', task_name)
        event_.wait()
    logger.debug('all tasks in schedule finished executing')


@event('manager.daemon.started')
def setup_scheduler(manager):
    """Configure and start apscheduler"""
    global scheduler
    if logger.level(manager.options.loglevel).no > logger.level('DEBUG').no:
        logging.getLogger('apscheduler').setLevel(logging.WARNING)
    # Since APScheduler runs in a separate thread, slower devices can sometimes get a DB lock, so use a separate db
    # for the jobs to avoid this
    db_filename = os.path.join(manager.config_base, 'db-%s-jobs.sqlite' % manager.config_name)
    # in case running on windows, needs double \\
    db_filename = db_filename.replace('\\', '\\\\')
    database_uri = 'sqlite:///%s' % db_filename
    jobstores = {'default': SQLAlchemyJobStore(url=database_uri)}
    # If job was meant to run within last day while daemon was shutdown, run it once when continuing
    job_defaults = {'coalesce': True, 'misfire_grace_time': 60 * 60 * 24}
    try:
        timezone = tzlocal.get_localzone()
        if timezone.zone == 'local':
            timezone = None
    except pytz.UnknownTimeZoneError:
        timezone = None
    except struct.error as e:
        # Hiding exception that may occur in tzfile.py seen in entware
        logger.warning('Hiding exception from tzlocal: {}', e)
        timezone = None
    if not timezone:
        # The default sqlalchemy jobstore does not work when there isn't a name for the local timezone.
        # Just fall back to utc in this case
        # FlexGet #2741, upstream ticket https://github.com/agronholm/apscheduler/issues/59
        logger.info(
            'Local timezone name could not be determined. Scheduler will display times in UTC for any log'
            'messages. To resolve this set up /etc/timezone with correct time zone name.'
        )
        timezone = pytz.utc
    scheduler = BackgroundScheduler(
        jobstores=jobstores, job_defaults=job_defaults, timezone=timezone
    )
    setup_jobs(manager)


@event('manager.config_updated')
def setup_jobs(manager):
    """Set up the jobs for apscheduler to run."""
    if not manager.is_daemon:
        return

    global scheduler_job_map
    scheduler_job_map = {}

    if 'schedules' not in manager.config:
        logger.info(
            'No schedules defined in config. Defaulting to run all tasks on a 1 hour interval.'
        )
    config = manager.config.get('schedules', True)
    if config is True:
        config = DEFAULT_SCHEDULES
    elif not config:  # Schedules are disabled with `schedules: no`
        if scheduler.running:
            logger.info('Shutting down scheduler')
            scheduler.shutdown()
        return
    if not scheduler.running:
        logger.info('Starting scheduler')
        scheduler.start(paused=True)
    existing_job_ids = [job.id for job in scheduler.get_jobs()]
    configured_job_ids = []
    for job_config in config:
        jid = job_id(job_config)
        configured_job_ids.append(jid)
        scheduler_job_map[id(job_config)] = jid
        if jid in existing_job_ids:
            continue
        if 'interval' in job_config:
            trigger, trigger_args = 'interval', job_config['interval']
        else:
            trigger, trigger_args = 'cron', job_config['schedule']
        tasks = job_config['tasks']
        if not isinstance(tasks, list):
            tasks = [tasks]
        name = ','.join(tasks)
        scheduler.add_job(
            run_job, args=(tasks,), id=jid, name=name, trigger=trigger, **trigger_args
        )
    # Remove jobs no longer in config
    for jid in existing_job_ids:
        if jid not in configured_job_ids:
            scheduler.remove_job(jid)
    scheduler.resume()


@event('manager.shutdown_requested')
def shutdown_requested(manager):
    if scheduler and scheduler.running:
        scheduler.shutdown(wait=True)


@event('manager.shutdown')
def stop_scheduler(manager):
    if scheduler and scheduler.running:
        scheduler.shutdown(wait=False)


@event('config.register')
def register_config():
    register_config_key('schedules', main_schema)
    register_schema('/schema/config/schedule', schedule_schema)



from urllib.parse import quote

from jinja2 import TemplateSyntaxError
from loguru import logger

from flexget import plugin
from flexget.components.sites.utils import normalize_unicode
from flexget.event import event

logger = logger.bind(name='search_rss')


class SearchRSS:
    """A generic search plugin that can use rss based search feeds. Configure it like rss
    plugin, but include {{{search_term}}} in the url where the search term should go."""

    schema = {'$ref': '/schema/plugin/rss'}

    def search(self, task, entry, config=None):
        from flexget.utils.template import environment

        search_strings = [
            quote(normalize_unicode(s).encode('utf-8'))
            for s in entry.get('search_strings', [entry['title']])
        ]
        rss_plugin = plugin.get_plugin_by_name('rss')
        entries = set()
        rss_config = rss_plugin.instance.build_config(config)
        try:
            template = environment.from_string(rss_config['url'])
        except TemplateSyntaxError as e:
            raise plugin.PluginError('Invalid jinja template as rss url: %s' % e)
        rss_config['all_entries'] = True
        for search_string in search_strings:
            rss_config['url'] = template.render({'search_term': search_string})
            # TODO: capture some other_fields to try to find seed/peer/content_size numbers?
            try:
                results = rss_plugin.phase_handlers['input'](task, rss_config)
            except plugin.PluginError as e:
                logger.error('Error attempting to get rss for {}: {}', rss_config['url'], e)
            else:
                entries.update(results)
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchRSS, 'search_rss', interfaces=['search'], api_ver=2)



from flask import jsonify, request, Response
from sqlalchemy.orm import Session

from flexget.api import APIResource, api
from flexget.api.app import BadRequest, base_message_schema, success_response
from flexget.db_schema import plugin_schemas, reset_schema

db_api = api.namespace('database', description='Manage Flexget DB')


class ObjectsContainer:
    plugin_list = {'type': 'array', 'items': {'type': 'string'}}

    database_input_object = {
        'type': 'object',
        'properties': {
            'operation': {'type': 'string', 'enum': ['cleanup', 'vacuum', 'plugin_reset']},
            'plugin_name': {'type': 'string'},
        },
        'required': ['operation'],
        'additionalProperties': False,
    }


plugins_schema = api.schema_model('plugins_list', ObjectsContainer.plugin_list)
input_schema = api.schema_model('db_schema', ObjectsContainer.database_input_object)


@db_api.route('/')
class DBOperation(APIResource):
    @api.validate(input_schema)
    @api.response(200, model=base_message_schema)
    def post(self, session: Session = None) -> Response:
        """Perform DB operations"""
        msg = ''
        data = request.json
        operation = data['operation']
        if operation == 'cleanup':
            self.manager.db_cleanup(force=True)
            msg = 'DB Cleanup finished'
        elif operation == 'vacuum':
            session.execute('VACUUM')
            session.commit()
            msg = 'DB VACUUM finished'
        elif operation == 'plugin_reset':
            plugin_name = data.get('plugin_name')
            if not plugin_name:
                raise BadRequest(
                    "'plugin_name' attribute must be used when trying to reset plugin"
                )
            try:
                reset_schema(plugin_name)
                msg = f'Plugin {plugin_name} DB reset was successful'
            except ValueError:
                raise BadRequest(f'The plugin {plugin_name} has no stored schema to reset')
        return success_response(msg)


@db_api.route('/plugins/')
class DBCleanup(APIResource):
    @api.response(200, model=plugins_schema)
    def get(self, session: Session = None) -> Response:
        """List resettable DB plugins"""
        return jsonify(sorted(list(plugin_schemas)))



import logging
import time
from functools import partial
from pathlib import Path, PurePath, PurePosixPath
from typing import Callable, List, Optional
from urllib.parse import quote, urljoin

from loguru import logger

from flexget import plugin
from flexget.entry import Entry

# retry configuration constants
RETRY_INTERVAL_SEC: int = 15
RETRY_STEP_SEC: int = 5

try:
    import pysftp

    logging.getLogger("paramiko").setLevel(logging.ERROR)
except ImportError:
    pysftp = None

NodeHandler = Callable[[str], None]

logger = logger.bind(name='sftp_client')


class SftpClient:
    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: Optional[str] = None,
        private_key: Optional[str] = None,
        private_key_pass: Optional[str] = None,
        connection_tries: int = 3,
    ):

        if not pysftp:
            raise plugin.DependencyError(
                issued_by='sftp_client',
                missing='pysftp',
                message='sftp client requires the pysftp Python module.',
            )

        self.host: str = host
        self.port: int = port
        self.username: str = username
        self.password: Optional[str] = password
        self.private_key: Optional[str] = private_key
        self.private_key_pass: Optional[str] = private_key_pass

        self.prefix: str = self._get_prefix()
        self._sftp: 'pysftp.Connection' = self._connect(connection_tries)
        self._handler_builder: HandlerBuilder = HandlerBuilder(
            self._sftp, self.prefix, self.private_key, self.private_key_pass
        )

    def list_directories(
        self, directories: List[str], recursive: bool, get_size: bool, files_only: bool
    ) -> List[Entry]:
        """
        Build a list of entries from a provided list of directories on an SFTP server
        :param directories: list of directories to generate entries for
        :param recursive: boolean indicating whether to list recursively
        :param get_size: boolean indicating whether to compute size for each node (potentially slow for directories)
        :param files_only: boolean indicating whether to exclude directories
        :return: a list of entries describing the contents of the provided directories
        """

        entries: List[Entry] = []

        dir_handler: NodeHandler = self._handler_builder.get_dir_handler(
            get_size, files_only, entries
        )
        file_handler: NodeHandler = self._handler_builder.get_file_handler(get_size, entries)
        unknown_handler: NodeHandler = self._handler_builder.get_unknown_handler()

        for directory in directories:
            try:
                self._sftp.walktree(
                    directory, file_handler, dir_handler, unknown_handler, recursive
                )
            except OSError as e:
                logger.warning('Failed to open {} ({})', directory, str(e))
                continue

        return entries

    def download(self, source: str, to: str, recursive: bool, delete_origin: bool) -> None:
        """
        Downloads the file specified in "source" to the destination specified in "to"
        :param source: path of the resource to download
        :param to: path of the directory to download to
        :param recursive: indicates whether to download the contents of "source" recursively
        :param delete_origin: indicates whether to delete the source resource upon download
        """

        dir_handler: NodeHandler = self._handler_builder.get_null_handler()
        unknown_handler: NodeHandler = self._handler_builder.get_unknown_handler()

        parsed_path: PurePosixPath = PurePosixPath(source)

        if not self.path_exists(source):
            raise SftpError(f'Remote path does not exist: {source}')

        if self.is_file(source):
            source_file: str = parsed_path.name
            source_dir: str = parsed_path.parent.as_posix()
            try:
                self._sftp.cwd(source_dir)
                self._download_file(to, delete_origin, source_file)
            except Exception as e:
                raise SftpError(f'Failed to download file {source} ({str(e)})')

            if delete_origin:
                self.remove_dir(source_dir)
        elif self.is_dir(source):
            base_path: str = parsed_path.joinpath('..').as_posix()
            dir_name: str = parsed_path.name
            handle_file: NodeHandler = partial(self._download_file, to, delete_origin)

            try:
                self._sftp.cwd(base_path)
                self._sftp.walktree(dir_name, handle_file, dir_handler, unknown_handler, recursive)
            except Exception as e:
                raise SftpError(f'Failed to download directory {source} ({str(e)})')

            if delete_origin:
                self.remove_dir(source)
        else:
            logger.warning('Skipping unknown file: {}', source)

    def upload(self, source: str, to: str) -> None:
        """
        Upload files or directories to an SFTP server
        :param source: file or directory to upload
        :param to: destination
        """
        if Path(source).is_dir():
            logger.verbose('Skipping directory {}', source)  # type: ignore
        else:
            self._upload_file(source, to)

    def remove_dir(self, path: str) -> None:
        """
        Remove a directory if it's empty
        :param path: directory to remove
        """
        if self._sftp.exists(path) and not self._sftp.listdir(path):
            logger.debug('Attempting to delete directory {}', path)
            try:
                self._sftp.rmdir(path)
            except Exception as e:
                logger.error('Failed to delete directory {} ({})', path, str(e))

    def remove_file(self, path: str) -> None:
        """
        Remove a file if it's empty
        :param path: file to remove
        """
        logger.debug('Deleting remote file {}', path)
        try:
            self._sftp.remove(path)
        except Exception as e:
            logger.error('Failed to delete file {} ({})', path, str(e))
            return

    def is_file(self, path: str) -> bool:
        """
        Check if the node at a given path is a file
        :param path: path to check
        :return: boolean indicating if the path is a file
        """
        return self._sftp.isfile(path)

    def is_dir(self, path: str) -> bool:
        """
        Check if the node at a given path is a directory
        :param path: path to check
        :return: boolean indicating if the path is a directory
        """
        return self._sftp.isdir(path)

    def path_exists(self, path: str) -> bool:
        """
        Check of a path exists

        :param path: Path to check
        :return: boolean indicating if the path exists
        """
        return self._sftp.lexists(path)

    def make_dirs(self, path: str) -> None:
        """
        Build directories

        :param path: path to build
        """

        if not self.path_exists(path):
            try:
                self._sftp.makedirs(path)
            except Exception as e:
                raise SftpError(f'Failed to create remote directory {path} ({str(e)})')

    def close(self) -> None:
        """
        Close the sftp connection
        """
        self._sftp.close()

    def set_socket_timeout(self, socket_timeout_sec):
        """
        Sets the SFTP client socket timeout
        :param socket_timeout_sec: Socket timeout in seconds
        """
        self._sftp.timeout = socket_timeout_sec

    def _connect(self, connection_tries: int) -> 'pysftp.Connection':

        tries: int = connection_tries
        retry_interval: int = RETRY_INTERVAL_SEC

        logger.debug('Connecting to {}', self.host)

        sftp: Optional['pysftp.Connection'] = None

        while not sftp:
            try:
                sftp = pysftp.Connection(
                    host=self.host,
                    username=self.username,
                    private_key=self.private_key,
                    password=self.password,
                    port=self.port,
                    private_key_pass=self.private_key_pass,
                )
                logger.verbose('Connected to {}', self.host)  # type: ignore
            except Exception as e:
                tries -= 1
                if not tries:
                    raise e
                else:
                    logger.debug('Caught exception: {}', e)
                    logger.warning(
                        'Failed to connect to {}; waiting {} seconds before retrying.',
                        self.host,
                        retry_interval,
                    )
                    time.sleep(retry_interval)
                    retry_interval += RETRY_STEP_SEC

        return sftp

    def _upload_file(self, source: str, to: str) -> None:
        if not Path(source).exists():
            logger.warning('File no longer exists:', source)
            return

        destination = self._get_upload_path(source, to)
        destination_url: str = urljoin(self.prefix, destination)

        if not self.path_exists(to):
            try:
                self.make_dirs(to)
            except Exception as e:
                raise SftpError(f'Failed to create remote directory {to} ({str(e)})')

        if not self.is_dir(to):
            raise SftpError(f'Not a directory: {to}')

        try:
            self._put_file(source, destination)
            logger.verbose('Successfully uploaded {} to {}', source, destination_url)  # type: ignore
        except OSError:
            raise SftpError(f'Remote directory does not exist: {to}')
        except Exception as e:
            raise SftpError(f'Failed to upload {source} ({str(e)})')

    def _download_file(self, destination: str, delete_origin: bool, source: str) -> None:

        destination_path: str = self._get_download_path(source, destination)
        destination_dir: str = Path(destination_path).parent.as_posix()

        if Path(destination_path).exists():
            logger.verbose(  # type: ignore
                'Skipping {} because destination file {} already exists.', source, destination_path
            )
            return

        Path(destination_dir).mkdir(parents=True, exist_ok=True)

        logger.verbose('Downloading file {} to {}', source, destination)  # type: ignore

        try:
            self._sftp.get(source, destination_path)
        except Exception as e:
            logger.error('Failed to download {} ({})', source, e)
            if Path(destination_path).exists():
                logger.debug('Removing partially downloaded file {}', destination_path)
                Path(destination_path).unlink()
            raise e

        if delete_origin:
            self.remove_file(source)

    def _put_file(self, source: str, destination: str) -> None:
        return self._sftp.put(source, destination)

    def _get_prefix(self) -> str:
        """
        Generate SFTP URL prefix
        """

        def get_login_string() -> str:
            if self.username and self.password:
                return f'{self.username}:{self.password}@'
            elif self.username:
                return f'{self.username}@'
            else:
                return ''

        def get_port_string() -> str:
            if self.port and self.port != 22:
                return f':{self.port}'
            else:
                return ''

        login_string = get_login_string()
        host = self.host
        port_string = get_port_string()

        return f'sftp://{login_string}{host}{port_string}/'

    @staticmethod
    def _get_download_path(path: str, destination: str) -> str:
        return PurePosixPath(destination).joinpath(Path(path)).as_posix()

    @staticmethod
    def _get_upload_path(source: str, to: str):
        basename: str = PurePath(source).name
        return PurePosixPath(to, basename).as_posix()


class SftpError(Exception):
    pass


class HandlerBuilder:
    """
    Class for building pysftp.Connection.walktree node handlers.

    :param sftp: A Connection object
    :param logger: a logger object
    :param url_prefix: SFTP URL prefix
    """

    def __init__(
        self,
        sftp: 'pysftp.Connection',
        url_prefix: str,
        private_key: Optional[str],
        private_key_pass: Optional[str],
    ):
        self._sftp = sftp
        self._prefix = url_prefix
        self._private_key = private_key
        self._private_key_pass = private_key_pass

    def get_file_handler(self, get_size: bool, entry_accumulator: list) -> NodeHandler:
        """
        Builds a file node handler suitable for use with pysftp.Connection.walktree

        :param get_size: boolean indicating whether to compute the for each file
        :param entry_accumulator: list to add entries to
        """
        return partial(
            Handlers.handle_file,
            self._sftp,
            self._prefix,
            get_size,
            self._private_key,
            self._private_key_pass,
            entry_accumulator,
        )

    def get_dir_handler(
        self, get_size: bool, files_only: bool, entry_accumulator: list
    ) -> NodeHandler:
        """
        Builds a file node handler suitable for use with pysftp.Connection.walktree

        :param get_size: boolean indicating whether to compute the for each file
        :param files_only: Boolean indicating whether to skip directories
        :param entry_accumulator: list to add entries to
        """
        return partial(
            Handlers.handle_directory,
            self._sftp,
            self._prefix,
            get_size,
            files_only,
            self._private_key,
            self._private_key_pass,
            entry_accumulator,
        )

    def get_unknown_handler(self) -> NodeHandler:
        """
        Builds an unknown node handler suitable for use with pysftp.Connection.walktree
        """
        return partial(Handlers.handle_unknown)

    def get_null_handler(self) -> NodeHandler:
        """
        Builds a noop node handler suitable for use with pysftp.Connection.walktree
        """
        return partial(Handlers.null_node_handler)


class Handlers:
    @classmethod
    def handle_file(
        cls,
        sftp: 'pysftp.Connection',
        prefix: str,
        get_size: bool,
        private_key: Optional[str],
        private_key_pass: Optional[str],
        entry_accumulator: List[Entry],
        path: str,
    ) -> None:
        """
        File node handler. Adds a file entry to entry_accumulator.

        :param sftp: A pysftp.Connection object
        :param logger: a logger object
        :param prefix: SFTP URL prefix
        :param get_size: boolean indicating whether to compute the size of each file
        :param private_key: private key path
        :param private_key_pass: private key password
        :param entry_accumulator: a list in which to store entries
        :param path: path to handle
        """
        size_handler = partial(cls._file_size, sftp)
        entry = cls._get_entry(
            sftp, prefix, size_handler, get_size, path, private_key, private_key_pass
        )
        entry_accumulator.append(entry)

    @classmethod
    def handle_directory(
        cls,
        sftp: 'pysftp.Connection',
        prefix: str,
        get_size: bool,
        files_only: bool,
        private_key: Optional[str],
        private_key_pass: Optional[str],
        entry_accumulator: List[Entry],
        path: str,
    ) -> None:
        """
        Directory node handler. Adds a directory entry to entry_accumulator.

        :param sftp: A pysftp.Connection object
        :param logger: a logger object
        :param prefix: SFTP URL prefix
        :param get_size: boolean indicating whether to compute the size of each directory
        :param files_only: Boolean indicating whether to skip directories
        :param entry_accumulator: a list in which to store entries
        :param private_key: private key path
        :param private_key_pass: private key password
        :param path: path to handle
        """
        if files_only:
            return

        dir_size: Callable[[str], int] = partial(cls._dir_size, sftp)
        entry: Entry = cls._get_entry(
            sftp, prefix, dir_size, get_size, path, private_key, private_key_pass
        )
        entry_accumulator.append(entry)

    @staticmethod
    def handle_unknown(path: str) -> None:
        """
        Handler for unknown nodes; logs a warning.

        :param logger: a logger object
        :param path: path to handle
        """
        logger.warning('Skipping unknown file: {}', path)

    @staticmethod
    def null_node_handler(path: str) -> None:
        """
        Generic noop node handler

        :param logger: a logger object
        :param path: path to handle
        :return:
        """
        logger.debug('null handler called  for {}', path)

    @staticmethod
    def _get_entry(
        sftp: 'pysftp.Connection',
        prefix: str,
        size_handler: Callable[[str], int],
        get_size,
        path: str,
        private_key: Optional[str],
        private_key_pass: Optional[str],
    ) -> Entry:

        url = urljoin(prefix, quote(sftp.normalize(path)))
        title = PurePosixPath(path).name

        entry = Entry(title, url)

        if get_size:
            try:
                size = size_handler(path)
            except Exception as e:
                logger.warning('Failed to get size for {} ({})', path, e)
                size = -1
            entry['content_size'] = size

        entry['private_key'] = private_key
        entry['private_key_pass'] = private_key_pass

        return entry

    @classmethod
    def _dir_size(cls, sftp: 'pysftp.Connection', path: str) -> int:
        sizes: List[int] = []

        size_accumulator = partial(cls._accumulate_file_size, sftp, sizes)
        sftp.walktree(path, size_accumulator, size_accumulator, size_accumulator, True)

        return sum(sizes)

    @classmethod
    def _accumulate_file_size(
        cls, sftp: 'pysftp.Connection', size_accumulator: List[int], path: str
    ) -> None:
        size_accumulator.append(cls._file_size(sftp, path))

    @staticmethod
    def _file_size(sftp: 'pysftp.Connection', path: str) -> int:
        """
        Helper function to get the size of a file node
        """
        return sftp.lstat(path).st_size



from flask import jsonify, request, Response
from flask_login import current_user
from sqlalchemy.orm import Session

from flexget.api import APIResource, api
from flexget.api.app import BadRequest, base_message_schema, success_response
from flexget.webserver import WeakPassword, change_password, generate_token

user_api = api.namespace('user', description='Manage user login credentials')


class ObjectsContainer:
    user_password_input = {
        'type': 'object',
        'properties': {'password': {'type': 'string'}},
        'required': ['password'],
        'additionalProperties': False,
    }

    user_token_response = {'type': 'object', 'properties': {'token': {'type': 'string'}}}


user_password_input_schema = api.schema_model(
    'user_password_input', ObjectsContainer.user_password_input
)
user_token_response_schema = api.schema_model(
    'user_token_response', ObjectsContainer.user_token_response
)


@user_api.route('/')
@api.doc('Change user password')
class UserManagementAPI(APIResource):
    @api.validate(model=user_password_input_schema, description='Password change schema')
    @api.response(BadRequest)
    @api.response(200, 'Success', model=base_message_schema)
    @api.doc(
        description='Change user password. A score of at least 3 is needed.'
        'See https://github.com/dropbox/zxcvbn for details'
    )
    def put(self, session: Session = None) -> Response:
        """ Change user password """
        user = current_user
        data = request.json
        try:
            change_password(username=user.name, password=data.get('password'), session=session)
        except WeakPassword as e:
            raise BadRequest(e.value)
        return success_response('Successfully changed user password')


@user_api.route('/token/')
@api.doc('Change user token')
class UserManagementTokenAPI(APIResource):
    @api.response(200, 'Successfully got user token', user_token_response_schema)
    @api.doc(description='Get current user token')
    def get(self, session: Session = None) -> Response:
        token = current_user.token
        return jsonify({'token': token})

    @api.response(200, 'Successfully changed user token', user_token_response_schema)
    @api.doc(description='Get new user token')
    def put(self, session: Session = None) -> Response:
        """ Change current user token """
        token = generate_token(username=current_user.name, session=session)
        return jsonify({'token': token})



"""
Utilities for handling RAR and ZIP archives

Provides wrapper archive and exception classes to simplify
archive extraction
"""


import os
import shutil
import zipfile

from loguru import logger

try:
    import rarfile
except ImportError:
    rarfile = None

logger = logger.bind(name='archive')


class ArchiveError(Exception):
    """Base exception for archive"""

    pass


class NeedRarFile(ArchiveError):
    """Exception to be raised when rarfile module is missing"""

    pass


class BadArchive(ArchiveError):
    """Wrapper exception for BadZipFile and BadRarFile"""

    pass


class NeedFirstVolume(ArchiveError):
    """Wrapper exception for rarfile.NeedFirstVolume"""

    pass


class PathError(ArchiveError):
    """Exception to be raised when an archive file doesn't exist"""

    pass


class FSError(ArchiveError):
    """Exception to be raised on OS/IO exceptions"""

    pass


class FileAlreadyExists(ArchiveError):
    """Exception to be raised when destination file already exists"""

    pass


def rarfile_set_tool_path(config):
    """
    Manually set the path of unrar executable if it can't be resolved from the
    PATH environment variable
    """
    unrar_tool = config['unrar_tool']

    if unrar_tool:
        if not rarfile:
            logger.error('rar_tool specified with no rarfile module installed.')
        else:
            rarfile.UNRAR_TOOL = unrar_tool
            logger.debug('Set RarFile.unrar_tool to: {}', unrar_tool)


def rarfile_set_path_sep(separator):
    """
    Set the path separator on rarfile module
    """
    if rarfile:
        rarfile.PATH_SEP = separator


def makepath(path):
    """Make directories as needed"""
    if not os.path.exists(path):
        logger.debug('Creating path: {}', path)
        os.makedirs(path)


class Archive:
    """
    Base archive class. Assumes an interface similar to
    zipfile.ZipFile or rarfile.RarFile
    """

    def __init__(self, archive_object, path):
        self.path = path

        self.archive = archive_object(self.path)

    def close(self):
        """Release open resources."""
        self.archive.close()

    def delete(self):
        """Delete the volumes that make up this archive"""
        volumes = self.volumes()
        self.close()

        try:
            for volume in volumes:
                os.remove(volume)
                logger.verbose('Deleted archive: {}', volume)
        except OSError as error:
            raise FSError(error)

    def volumes(self):
        """Returns the list of volumes that comprise this archive"""
        return [self.path]

    def infolist(self):
        """Returns a list of info objects describing the contents of this archive"""
        infolist = []

        for info in self.archive.infolist():
            try:
                archive_info = ArchiveInfo(info)
                infolist.append(archive_info)
            except ValueError as e:
                logger.debug(e)

        return infolist

    def open(self, member):
        """Returns file-like object from where the data of a member file can be read."""
        return self.archive.open(member)

    def extract_file(self, member, destination):
        """Extract a member file to the specified destination"""
        try:
            with self.open(member) as source:
                with open(destination, 'wb') as target:
                    shutil.copyfileobj(source, target)
        except OSError as error:
            raise FSError(error)

        logger.verbose('Extracted: {}', member)


class RarArchive(Archive):
    """
    Wrapper class for rarfile.RarFile
    """

    def __init__(self, path):
        if not rarfile:
            raise NeedRarFile('Python module rarfile needed to handle RAR archives')

        try:
            super().__init__(rarfile.RarFile, path)
        except rarfile.BadRarFile as error:
            raise BadArchive(error)
        except rarfile.NeedFirstVolume as error:
            raise NeedFirstVolume(error)
        except rarfile.Error as error:
            raise ArchiveError(error)

    def volumes(self):
        """Returns the list of volumes that comprise this archive"""
        return self.archive.volumelist()

    def open(self, member):
        """Returns file-like object from where the data of a member file can be read."""
        try:
            return super().open(member)
        except rarfile.Error as error:
            raise ArchiveError(error)


class ZipArchive(Archive):
    """
    Wrapper class for zipfile.ZipFile
    """

    def __init__(self, path):
        try:
            super().__init__(zipfile.ZipFile, path)
        except zipfile.BadZipfile as error:
            raise BadArchive(error)

    def open(self, member):
        """Returns file-like object from where the data of a member file can be read."""
        try:
            return super().open(member)
        except zipfile.BadZipfile as error:
            raise ArchiveError(error)


class ArchiveInfo:
    """Wrapper class for  archive info objects"""

    def __init__(self, info):
        self.info = info
        self.path = info.filename
        self.filename = os.path.basename(self.path)

        if self._is_dir():
            raise ValueError('Appears to be a directory: %s' % self.path)

    def _is_dir(self):
        """Indicates if info object looks to be a directory"""

        if hasattr(self.info, 'isdir'):
            return self.info.isdir()
        else:
            return not self.filename

    def extract(self, archive, destination):
        """Extract ArchiveInfo object to the specified destination"""
        dest_dir = os.path.dirname(destination)

        if os.path.exists(destination):
            raise FileAlreadyExists('File already exists: %s' % destination)

        logger.debug('Creating path: {}', dest_dir)
        makepath(dest_dir)

        try:
            archive.extract_file(self.info, destination)
        except Exception as error:
            if os.path.exists(destination):
                logger.debug('Cleaning up partially extracted file: {}', destination)
                os.remove(destination)

            raise error


def open_archive(archive_path):
    """
    Returns the appropriate archive object
    """

    archive = None

    if not os.path.exists(archive_path):
        raise PathError('Path doesn\'t exist')

    if zipfile.is_zipfile(archive_path):
        archive = ZipArchive(archive_path)
        logger.debug('Successfully opened ZIP: {}', archive_path)
    elif rarfile and rarfile.is_rarfile(archive_path):
        archive = RarArchive(archive_path)
        logger.debug('Successfully opened RAR: {}', archive_path)
    else:
        if not rarfile:
            logger.warning('Rarfile module not installed; unable to handle RAR archives.')

    return archive


def is_archive(path):
    """
    Attempts to open an entry as an archive; returns True on success, False on failure.
    """

    archive = None

    try:
        archive = open_archive(path)
        if archive:
            archive.close()
            return True
    except (OSError, ArchiveError) as error:
        logger.debug('Failed to open file as archive: {} ({})', path, error)

    return False



import re

import feedparser
from loguru import logger
from requests.auth import AuthBase

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException

logger = logger.bind(name='apple_trailers')


class AppleTrailers:
    """
    Adds support for Apple.com movie trailers.

    Configuration:
    quality: Set the desired resolution - 480p, 720p or 1080p. default '720p'
    genres:  List of genres used to filter the entries. If set, the
    trailer must match at least one listed genre to be accepted. Genres
    that can be used: Action and Adventure, Comedy, Documentary, Drama,
    Family, Fantasy, Foreign, Horror, Musical, Romance, Science Fiction,
    Thriller. default '' (all)

    apple_trailers:
      quality: 720p
      genres: ['Action and Adventure']

    Alternatively, a simpler configuration format can be used. This uses
    the default genre filter, all:

    apple_trailers: 720p

    This plugin adds the following fields to the entry:
      movie_name, movie_year, genres, apple_trailers_name, movie_studio
    movie_name: Name of the movie
    movie_year: Year the movie was/will be released
    genres: Comma-separated list of genres that apply to the movie
    apple_trailers_name: Contains the Apple-supplied name of the clip,
    such as 'Clip 2', 'Trailer', 'Winter Olympic Preview'
    movie_studio: Name of the studio that makes the movie
    """

    movie_data_url = 'http://trailers.apple.com/trailers/feeds/data/'
    rss_url = 'http://trailers.apple.com/trailers/home/rss/newtrailers.rss'
    qualities = {'480p': 'sd', '720p': 'hd720', '1080p': 'hd1080'}

    schema = {
        'oneOf': [
            {
                'type': 'object',
                'properties': {
                    'quality': {
                        'type': 'string',
                        'enum': list(qualities.keys()),
                        'default': '720p',
                    },
                    'genres': {'type': 'array', 'items': {'type': 'string'}},
                },
                'additionalProperties': False,
            },
            {'title': 'justquality', 'type': 'string', 'enum': list(qualities.keys())},
        ]
    }

    def broken(self, error_message):
        raise plugin.PluginError('Plugin is most likely broken. Got: %s' % error_message)

    @plugin.priority(127)
    @cached('apple_trailers')
    def on_task_input(self, task, config):
        # Turn simple config into full config
        if isinstance(config, str):
            config = {'quality': config}

        try:
            r = task.requests.get(self.rss_url)
        except RequestException as e:
            raise plugin.PluginError('Retrieving Apple Trailers RSS feed failed: %s' % e)

        rss = feedparser.parse(r.content)

        if rss.get('bozo_exception', False):
            raise plugin.PluginError('Got bozo_exception (bad feed)')

        filmid_regex = re.compile(r'(FilmId\s*\=\s*\')(\d+)(?=\')')
        studio_regex = re.compile(r'(?:[0-9]*\s*)(.+)')
        # use the following dict to save json object in case multiple trailers have been released for the same movie
        # no need to do multiple requests for the same thing!
        trailers = {}
        entries = []
        for item in rss.entries:
            entry = Entry()
            movie_url = item['link']
            entry['title'] = item['title']
            entry['movie_name'], entry['apple_trailers_name'] = entry['title'].split(' - ', 1)
            if not trailers.get(movie_url):
                try:
                    movie_page = task.requests.get(movie_url).text
                    match = filmid_regex.search(movie_page)
                    if match:
                        json_url = self.movie_data_url + match.group(2) + '.json'
                        movie_data = task.requests.get(json_url).json()

                        trailers[movie_url] = {'json_url': json_url, 'json': movie_data}
                    else:
                        self.broken('FilmId not found for {0}'.format(entry['movie_name']))

                except RequestException as e:
                    logger.error('Failed to get trailer {}: {}', entry['title'], e.args[0])
                    continue
            else:
                movie_data = trailers[movie_url]['json']
            genres = {genre.get('name') for genre in movie_data.get('details').get('genres')}
            config_genres = set(config.get('genres', []))
            if genres and config_genres and not set.intersection(config_genres, genres):
                logger.debug('Config genre(s) do not match movie genre(s)')
                continue

            desired_quality = config['quality']
            # find the trailer url
            for clip in movie_data.get('clips'):
                if clip.get('title') == entry['apple_trailers_name']:
                    try:
                        trailer_url = clip['versions']['enus']['sizes'][
                            self.qualities[desired_quality]
                        ]
                        src = trailer_url.get('src')
                        src_alt = trailer_url.get('srcAlt')
                        # .mov tends to be a streaming video file, but the real video file is the same url, but
                        # they prepend 'h' to the quality
                        if src.split('.')[-1] == 'mov':
                            entry['url'] = src.replace(desired_quality, 'h' + desired_quality)
                        elif src_alt.split('.')[-1] == 'mov':
                            entry['url'] = src_alt.replace(desired_quality, 'h' + desired_quality)
                        else:
                            continue  # just continue until we reach the else part of the for-else
                        break
                    except KeyError as e:
                        self.broken(e.args[0])
            else:
                logger.error('Trailer "{}" not found', entry['apple_trailers_name'])
                continue

            # set some entry fields if present
            # studio is usually also the copyright holder
            studio = studio_regex.match(movie_data.get('page').get('copyright'))
            if studio:
                entry['movie_studio'] = studio.group(1)

            release_date = movie_data.get('page').get('release_date')
            if release_date:
                entry['release_date'] = release_date

            if genres:
                entry['genres'] = ', '.join(list(genres))

            # set the correct header without modifying the task.requests obj
            entry['download_auth'] = AppleTrailersHeader()
            entries.append(entry)

        return entries


class AppleTrailersHeader(AuthBase):
    def __call__(self, request):
        request.headers['User-Agent'] = 'QuickTime/7.7'
        return request


@event('plugin.register')
def register_plugin():
    plugin.register(AppleTrailers, 'apple_trailers', api_ver=2)



import math
import os.path
import re
from pathlib import Path

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='metanfo_csize')

SIZE_RE = re.compile(r'Size[^\d]{0,7}(\d*\.?\d+).{0,5}(MB|GB)', re.IGNORECASE)


class MetainfoContentSize:
    """
    Utility:

    Check if content size is mentioned in description and set content_size attribute for entries if it is.
    Also sets content_size for entries with local files from input_listdir.
    """

    schema = {'type': 'boolean', 'default': False}

    def on_task_metainfo(self, task, config):
        # check if disabled (value set to false)
        if config is False:
            return

        count = 0
        for entry in task.entries:
            if entry.get('content_size'):
                # Don't override if already set
                logger.trace(
                    'skipping content size check because it is already set for {!r}',
                    entry['title'],
                )
                continue
            # Try to parse size from description
            match = False
            description = entry.get('description', '')
            if isinstance(description, str):
                match = SIZE_RE.search(description)
            if match:
                try:
                    amount = float(match.group(1).replace(',', '.'))
                except Exception:
                    logger.error(
                        'BUG: Unable to convert {} into float ({})', match.group(1), entry['title']
                    )
                    continue
                unit = match.group(2).lower()
                count += 1
                if unit == 'gb':
                    amount = math.ceil(amount * 1024)
                logger.trace('setting content size to {}', amount)
                entry['content_size'] = int(amount)
                continue
            # If this entry has a local file, (it was added by filesystem plugin) grab the size.
            elif entry.get('location'):
                # If it is a .torrent or .nzb, don't bother getting the size as it will not be the content's size
                location = entry['location']
                if isinstance(location, str):
                    location = Path(location)
                if location.suffix in ('.nzb', '.torrent'):
                    continue
                try:
                    if location.is_file():
                        amount = os.path.getsize(entry['location'])
                        amount = int(amount / (1024 * 1024))
                        logger.trace('setting content size to {}', amount)
                        entry['content_size'] = amount
                        continue
                except OSError as e:
                    # is_file can throw OSError for invalid paths (at least on windows)
                    continue

        if count:
            logger.debug('Found content size information from {} entries', count)


@event('plugin.register')
def register_plugin():
    plugin.register(MetainfoContentSize, 'metainfo_content_size', builtin=True, api_ver=2)



import re
from urllib.parse import urlencode

from loguru import logger
from requests import Request, RequestException
from requests.utils import cookiejar_from_dict, dict_from_cookiejar

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginError

logger = logger.bind(name='wordpress_auth')


def construct_request(url, username='', password='', redirect='/wp-admin/'):
    headers = {
        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) '
        'Chrome/50.0.2661.102 Safari/537.36',
        'Content-Type': 'application/x-www-form-urlencoded',
        'DNT': '1',
    }
    data = {
        'log': username,
        'pwd': password,
        'wp-submit': 'Log In',
        'testcookie': '1',
        'redirect_to': redirect,
    }
    return Request(
        method='POST', url=url, headers=headers, data=urlencode(data).encode('UTF-8')
    ).prepare()


def collect_cookies(response):
    cookies = dict_from_cookiejar(response.cookies)
    for h_resp in response.history:
        cookies.update(dict_from_cookiejar(h_resp.cookies))
    return cookiejar_from_dict(cookies)


def get_valid_cookies(cookies):
    def is_wp_cookie(key):
        return re.match(r'(wordpress|wp)(?!_*test)[A-z0-9]*', key, re.IGNORECASE)

    valid_cookies = {key: value for key, value in cookies.items() if is_wp_cookie(key)}
    return cookiejar_from_dict(valid_cookies)


class PluginWordPress:
    """
    Supports accessing feeds and media that require wordpress account credentials
    Usage:

    wordpress_auth:
      url: 'your wordpress blog login page (ex http://example.org/wp-login.php)'
      username: 'your username'
      password: 'your password'
    """

    schema = {
        'type': 'object',
        'properties': {
            'url': {'type': 'string', 'oneOf': [{'format': 'url'}]},
            'username': {'type': 'string', 'default': ''},
            'password': {'type': 'string', 'default': ''},
        },
        'required': ['url'],
        'additionalProperties': False,
    }

    @plugin.priority(135)
    def on_task_start(self, task, config):
        url = config['url']
        username = config['username']
        password = config['password']
        try:
            response = task.requests.send(
                construct_request(url, username=username, password=password)
            )
            if not response.ok:
                raise RequestException(str(response))
            cookies = collect_cookies(response)
            if len(get_valid_cookies(cookies)) < 1:
                raise RequestException(
                    'No recognized WordPress cookies found. Perhaps username/password is invalid?'
                )
            task.requests.add_cookiejar(cookies)

        except RequestException as err:
            logger.error('{}', err)
            raise PluginError('WordPress Authentication at %s failed' % (url,))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginWordPress, 'wordpress_auth', api_ver=2)



import os

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='proxy')

PROTOCOLS = ['http', 'https']


class Proxy:
    """Adds a proxy to the requests session."""

    schema = {
        'oneOf': [
            {'type': 'string', 'format': 'url'},
            {
                'type': 'object',
                'properties': dict(
                    (prot, {'type': 'string', 'format': 'url'}) for prot in PROTOCOLS
                ),
                'additionalProperties': False,
            },
        ]
    }

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_prepare(self, task, config):
        if not config:
            # If no configuration is provided, see if there are any proxy env variables
            proxies = {}
            for prot in PROTOCOLS:
                if os.environ.get(prot + '_proxy'):
                    proxies[prot] = os.environ[prot + '_proxy']
            if not proxies:
                # If there were no environment variables set, do nothing
                return
        elif isinstance(config, dict):
            proxies = config
        else:
            # Map all protocols to the configured proxy
            proxies = dict((prot, config) for prot in PROTOCOLS)
        logger.verbose('Setting proxy to {}', proxies)
        task.requests.proxies = proxies


@event('plugin.register')
def register_plugin():
    plugin.register(Proxy, 'proxy', builtin=True, api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

from . import db

logger = logger.bind(name='series_forget')


class OutputSeriesRemove:
    schema = {'type': 'boolean'}

    def on_task_output(self, task, config):
        if not config:
            return
        for entry in task.accepted:
            if 'series_name' in entry and 'series_id' in entry:
                try:
                    db.remove_series_entity(entry['series_name'], entry['series_id'])
                    logger.info(
                        'Removed episode `{}` from series `{}` download history.',
                        entry['series_id'],
                        entry['series_name'],
                    )
                except ValueError:
                    logger.debug(
                        'Series ({}) or id ({}) unknown.', entry['series_name'], entry['series_id']
                    )


@event('plugin.register')
def register_plugin():
    plugin.register(OutputSeriesRemove, 'series_remove', api_ver=2)



import re
from urllib.parse import urlencode, urlparse

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.components.sites.utils import normalize_unicode, torrent_availability
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='piratebay')

URL = 'https://apibay.org'  # for TOR: http://piratebayztemzmv.onion

CATEGORIES = {
    'all': 0,
    'audio': 100,
    'music': 101,
    'video': 200,
    'movies': 201,
    'tv': 205,
    'highres movies': 207,
    'highres tv': 208,
    'comics': 602,
}

# try to maintain compatibility with old config
SORT = {
    'default': 'torrent_availability',
    'date': 'torrent_timestamp',
    'size': 'content_size',
    'seeds': 'torrent_seeds',
    'leechers': 'torrent_leeches',
}

# default trackers for magnet links
TRACKERS = [
    'udp://tracker.coppersurfer.tk:6969/announce',
    'udp://9.rarbg.to:2920/announce',
    'udp://tracker.opentrackr.org:1337',
    'udp://tracker.internetwarriors.net:1337/announce',
    'udp://tracker.leechers-paradise.org:6969/announce',
    'udp://tracker.coppersurfer.tk:6969/announce',
    'udp://tracker.pirateparty.gr:6969/announce',
    'udp://tracker.cyberia.is:6969/announce',
]


class UrlRewritePirateBay:
    """PirateBay urlrewriter."""

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'url': {'type': 'string', 'default': URL, 'format': 'url'},
                    'category': {
                        'oneOf': [
                            {'type': 'string', 'enum': list(CATEGORIES)},
                            {'type': 'integer'},
                        ]
                    },
                    'sort_by': {'type': 'string', 'enum': list(SORT)},
                    'sort_reverse': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        ]
    }

    url = None

    def __init__(self):
        self.set_urls(URL)

    def on_task_start(self, task, config=None):
        if not isinstance(config, dict):
            config = {}
        self.set_urls(config.get('url', URL))

    def set_urls(self, url):
        url = url.rstrip('/')
        if self.url != url:
            self.url = url
            parsed_url = urlparse(url)
            escaped_url_scheme = re.escape(parsed_url.scheme)
            escaped_url_netloc = re.escape(parsed_url.netloc)
            # most api mirrors share the same domain with frontend, except thepiratebay.org which uses apibay.org
            # some api mirrors might contain path, for ex https://pirateproxy.surf/api?url=/q.php?q=ubuntu&cat=0
            # valid URLs are https://piratebay.org/description.php?id=\d+ and https://apibay.org/description.php?id=\d+
            self.url_match = re.compile(
                fr'^{escaped_url_scheme}://(?:thepiratebay\.org(?:\:\d+)?|{escaped_url_netloc})/description\.php\?id=(\d+)$'
            )
            self.url_search = re.compile(
                fr'^(?:https?://thepiratebay\.org(?:\:\d+)?|{escaped_url_scheme}://{escaped_url_netloc})/search\.php\?q=.*$'
            )

    # urlrewriter API
    def url_rewritable(self, task, entry):
        url = entry['url']
        if url.endswith('.torrent'):
            return False
        return any((self.url_match.match(url), self.url_search.match(url)))

    # urlrewriter API
    def url_rewrite(self, task, entry):
        if 'url' not in entry:
            logger.error("Didn't actually get a URL...")
        else:
            logger.debug('Got the URL: {}', entry['url'])
        if self.url_search.match(entry['url']):
            # use search
            results = self.search(task, entry)
            if not results:
                raise UrlRewritingError("No search results found")
            # TODO: Close matching was taken out of search methods, this may need to be fixed to be more picky
            entry['url'] = results[0]['url']
        else:
            torrent_id = self.url_match.match(entry['url']).group(1)
            url = f"{self.url}/t.php?id={torrent_id}"
            logger.debug('Getting info for torrent ID {}', torrent_id)
            json_result = task.requests.get(url).json()
            if json_result['id'] == '0':
                raise UrlRewritingError(f"Torrent with ID {torrent_id} does not exist.")
            entry['url'] = self.info_hash_to_magnet(json_result['info_hash'], json_result['name'])

    @plugin.internet(logger)
    def search(self, task, entry, config=None):
        """
        Search for name from piratebay.
        """
        if not isinstance(config, dict):
            config = {}
        self.set_urls(config.get('url', URL))
        sort = SORT.get(config.get('sort_by', 'default'))
        sort_reverse = bool(config.get('sort_reverse', 'True'))
        if isinstance(config.get('category'), int):
            category = config['category']
        else:
            category = CATEGORIES.get(config.get('category', 'all'))

        entries = set()
        for search_string in entry.get('search_strings', [entry['title']]):
            # query = normalize_unicode(search_string)
            params = {'q': search_string, 'cat': category}
            url = f"{self.url}/q.php"
            json_results = task.requests.get(url, params=params).json()
            if not json_results:
                raise plugin.PluginError("Error while searching piratebay for %s.", search_string)
            for result in json_results:
                if result['id'] == '0':
                    # JSON when no results found: [{ "id":"0", "name":"No results returned", ... }]
                    break
                entry = self.json_to_entry(result)
                entries.add(entry)

        return sorted(entries, reverse=sort_reverse, key=lambda x: x.get(sort))

    # convert an info_hash string to a magnet uri
    @staticmethod
    def info_hash_to_magnet(info_hash: str, name: str) -> str:
        magnet = {'xt': f"urn:btih:{info_hash}", 'dn': name, 'tr': TRACKERS}
        magnet_qs = urlencode(magnet, doseq=True, safe=':')
        magnet_uri = f"magnet:?{magnet_qs}"
        return magnet_uri

    # convert a single json result to an Entry
    def json_to_entry(self, json_result: dict) -> Entry:
        entry = Entry()
        entry['title'] = json_result['name']
        entry['torrent_seeds'] = int(json_result['seeders'])
        entry['torrent_leeches'] = int(json_result['leechers'])
        entry['torrent_timestamp'] = int(json_result['added'])  # custom field for sorting by date
        entry['torrent_availability'] = torrent_availability(
            entry['torrent_seeds'], entry['torrent_leeches']
        )
        entry['content_size'] = int(
            round(int(json_result['size']) / (1024 * 1024))
        )  # content_size is in MiB
        entry['torrent_info_hash'] = json_result['info_hash']
        entry['url'] = self.info_hash_to_magnet(json_result['info_hash'], json_result['name'])
        return entry


@event('plugin.register')
def register_plugin():
    plugin.register(
        UrlRewritePirateBay, 'piratebay', interfaces=['urlrewriter', 'search', 'task'], api_ver=2
    )



import ntpath
import re
import sys
from typing import Optional

os_mode = None  # Can be 'windows', 'mac', 'linux' or None. None will auto-detect os.
# Replacement order is important, don't use dicts to store
platform_replaces = {
    'windows': [
        ['[:*?"<>| ]+', ' '],  # Turn illegal characters into a space
        [r'[\.\s]+([/\\]|$)', r'\1'],
    ],  # Dots cannot end file or directory names
    'mac': [['[: ]+', ' ']],  # Only colon is illegal here
    'linux': [],  # No illegal chars
}


def pathscrub(dirty_path: str, os: Optional[str] = None, filename: bool = False) -> str:
    """
    Strips illegal characters for a given os from a path.

    :param dirty_path: Path to be scrubbed.
    :param os: Defines which os mode should be used, can be 'windows', 'mac', 'linux', or None to auto-detect
    :param filename: If this is True, path separators will be replaced with '-'
    :return: A valid path.
    """

    # See if global os_mode has been defined by pathscrub plugin
    if os_mode and not os:
        os = os_mode

    if not os:
        # If os is not defined, try to detect appropriate
        drive, path = ntpath.splitdrive(dirty_path)
        if sys.platform.startswith('win') or drive:
            os = 'windows'
        elif sys.platform.startswith('darwin'):
            os = 'mac'
        else:
            os = 'linux'
    replaces = platform_replaces[os]

    # Make sure not to mess with windows drive specifications
    drive, path = ntpath.splitdrive(dirty_path)

    if filename:
        path = path.replace('/', ' ').replace('\\', ' ')
    # Remove spaces surrounding path components
    path = '/'.join(comp.strip() for comp in path.split('/'))
    if os == 'windows':
        path = '\\'.join(comp.strip() for comp in path.split('\\'))
    for search, replace in replaces:
        path = re.sub(search, replace, path)
    path = path.strip()
    # If we stripped everything from a filename, complain
    if filename and dirty_path and not path:
        raise ValueError(
            'Nothing was left after stripping invalid characters from path `%s`!' % dirty_path
        )
    return drive + path



from collections.abc import MutableSet

from loguru import logger
from sqlalchemy import or_
from sqlalchemy.sql.elements import and_

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session

from . import db

plugin_name = 'pending_list'
logger = logger.bind(name=plugin_name)


class PendingListSet(MutableSet):
    def _db_list(self, session):
        return (
            session.query(db.PendingListList)
            .filter(db.PendingListList.name == self.list_name)
            .first()
        )

    def __init__(self, config):
        if not isinstance(config, dict):
            config = {'list_name': config}
        self.list_name = config.get('list_name')

        self.filter_approved = True
        if config.get('include') == "pending":
            self.filter_approved = False
        if config.get('include') == "all":
            self.filter_approved = None

        with Session() as session:
            if not self._db_list(session):
                session.add(db.PendingListList(name=self.list_name))

    def _entry_query(self, session, entry, approved=None):
        query = (
            session.query(db.PendingListEntry)
            .filter(db.PendingListEntry.list_id == self._db_list(session).id)
            .filter(
                or_(
                    db.PendingListEntry.title == entry['title'],
                    and_(
                        db.PendingListEntry.original_url,
                        db.PendingListEntry.original_url == entry['original_url'],
                    ),
                )
            )
        )

        if approved is not None:
            query = query.filter(db.PendingListEntry.approved == approved)

        return query.first()

    def __iter__(self):
        with Session() as session:
            query = self._db_list(session).entries.order_by(db.PendingListEntry.added.desc())
            if self.filter_approved is not None:
                query = query.filter(db.PendingListEntry.approved == self.filter_approved)
            for e in query.all():
                logger.debug('returning {}', e.entry)
                yield e.entry

    def __contains__(self, entry):
        with Session() as session:
            return self._entry_query(session, entry) is not None

    def __len__(self):
        with Session() as session:
            return self._db_list(session).entries.count()

    def discard(self, entry):
        with Session() as session:
            db_entry = self._entry_query(session=session, entry=entry)
            if db_entry:
                logger.debug('deleting entry {}', db_entry)
                session.delete(db_entry)

    def add(self, entry):
        with Session() as session:
            stored_entry = self._entry_query(session, entry)
            if stored_entry:
                # Refresh all the fields if we already have this entry
                logger.debug('refreshing entry {}', entry)
                stored_entry.entry = entry
            else:
                logger.debug('adding entry {} to list {}', entry, self._db_list(session).name)
                stored_entry = db.PendingListEntry(
                    entry=entry, pending_list_id=self._db_list(session).id
                )
            session.add(stored_entry)

    @property
    def immutable(self):
        return False

    def _from_iterable(self, it):
        return set(it)

    @property
    def online(self):
        """Set the online status of the plugin, online plugin should be treated differently in certain situations,
        like test mode"""
        return False

    def get(self, entry):
        with Session() as session:
            match = self._entry_query(session=session, entry=entry, approved=self.filter_approved)
            return Entry(match.entry) if match else None


class PendingList:
    schema = {
        'oneOf': [
            {'type': 'string'},
            {
                'type': 'object',
                'properties': {
                    'list_name': {'type': 'string'},
                    'include': {
                        'type': 'string',
                        'enum': ['pending', 'approved', 'all'],
                        'default': 'approved',
                    },
                },
                'required': ['list_name'],
                'additionalProperties': False,
            },
        ]
    }

    @staticmethod
    def get_list(config):
        return PendingListSet(config)

    def on_task_input(self, task, config):
        return list(PendingListSet(config))


@event('plugin.register')
def register_plugin():
    plugin.register(PendingList, plugin_name, api_ver=2, interfaces=['task', 'list'])



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='disable_phases')


class PluginDisablePhases:
    """Disables phases from task execution.

    Mainly meant for advanced users and development.

    Example:

    disable_phases:
      - download
    """

    @property
    def schema(self):
        return {'type': 'array', 'items': {'type': 'string', 'enum': plugin.task_phases}}

    def on_task_start(self, task, config):
        list(map(task.disable_phase, config))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginDisablePhases, 'disable_phases', api_ver=2)



import re

from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException
from flexget.utils.soup import get_soup

logger = logger.bind(name='horriblesubs')


class HorribleSubs:
    """
    Give latest horriblesubs releases
    """

    schema = {'type': 'boolean'}

    @staticmethod
    def horrible_get_downloads(requests, title, page_url):
        entries = []

        try:
            soup = get_soup(requests.get(page_url).content)
        except RequestException as e:
            logger.error('HorribleSubs request failed: {}', e)
            return entries
        for div in soup.findAll('div', attrs={'class': 'rls-link'}):
            ttitle = '{0} [{1}]'.format(title, re.sub(r'.*-', '', div['id']))
            urls = []
            for url in div.findAll('a'):
                # skip non torrent based links
                if (
                    'hs-ddl-link' in url.parent.attrs['class']
                    or 'hs-xdcc-link' in url.parent.attrs['class']
                ):
                    continue
                logger.debug('Found url `{}`', url)
                urls.append(url.attrs['href'])
            # move magnets to last, a bit hacky
            for url in urls[:]:
                if url.startswith('magnet'):
                    urls.remove(url)
                    urls.append(url)
            entries.append(Entry(title=ttitle, url=urls[0], urls=urls))
        return entries

    @staticmethod
    def horrible_entries(requests, page_url):
        entries = []

        try:
            soup = get_soup(requests.get(page_url).content)
        except RequestException as e:
            logger.error('HorribleSubs request failed: {}', e)
            return entries

        for li_label in soup.findAll('li'):
            title = '[HorribleSubs] {0}{1}'.format(
                str(li_label.find('span').next_sibling), str(li_label.find('strong').text)
            )
            logger.debug('Found title `{}`', title)
            url = li_label.find('a')['href']
            episode = re.sub(r'.*#', '', url)
            # Get show ID
            try:
                soup = get_soup(requests.get('https://horriblesubs.info/{0}'.format(url)).content)
            except RequestException as e:
                logger.error('HorribleSubs request failed: {}', e)
                return entries
            show_id = re.sub(r'[^0-9]', '', soup(text=re.compile('hs_showid'))[0])
            entries = HorribleSubs.horrible_get_downloads(
                requests,
                title,
                'https://horriblesubs.info/api.php?method=getshows&type=show&mode=filter&showid={0}&value={1}'.format(
                    show_id, episode
                ),
            )
        return entries

    @cached('horriblesubs')
    def on_task_input(self, task, config):
        if not config:
            return
        return HorribleSubs.horrible_entries(
            task.requests, 'https://horriblesubs.info/api.php?method=getlatest'
        )

    # Search API method
    def search(self, task, entry, config):
        if not config:
            return
        entries = []
        for search_string in entry.get('search_strings', [entry['title']]):
            logger.debug('Searching `{}`', search_string)
            results = HorribleSubs.horrible_entries(
                task.requests,
                'https://horriblesubs.info/api.php?method=search&value={0}'.format(search_string),
            )
            entries.extend(results)
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(HorribleSubs, 'horriblesubs', interfaces=['task', 'search'], api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='accept_all')


class FilterAcceptAll:
    """
    Just accepts all entries.

    Example::

      accept_all: true
    """

    schema = {'type': 'boolean'}

    def on_task_filter(self, task, config):
        if config:
            for entry in task.entries:
                entry.accept()


@event('plugin.register')
def register_plugin():
    plugin.register(FilterAcceptAll, 'accept_all', api_ver=2)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.manager import Session

logger = logger.bind(name='tvmaze_lookup')


class PluginTVMazeLookup:
    """Retrieves tvmaze information for entries. Uses series_name,
    series_season, series_episode from series plugin.

    Example:

    tvmaze_lookup: yes

    Primarily used for passing tvmaze information to other plugins.

    This information is provided (via entry):

    series info:
    tvmaze_series_name
    tvmaze_series_year
    tvdb_id
    tvrage_id
    tvmaze_series_id
    tvmaze_series_show_id
    tvmaze_series_tvrage
    tvmaze_series_runtime
    tvmaze_series_premiered
    tvmaze_series_airdays
    tvmaze_series_weight
    tvmaze_series_update_date
    tvmaze_series_language
    tvmaze_series_original_image
    tvmaze_series_medium_image
    tvmaze_series_summary
    tvmaze_series_webchannel
    tvmaze_series_show_type
    tvmaze_genres
    tvmaze_series_network
    tvmaze_series_url
    tvmaze_series_status
    tvmaze_series_rating
    tvmaze_series_episodes

    episode info:
    tvmaze_episode_name
    tvmaze_season
    tvmaze_episode
    tvmaze_episode_id
    tvmaze_episode_airdate
    tvmaze_episode_url
    tvmaze_episode_original_image
    tvmaze_episode_medium_image
    tvmaze_episode_airstamp
    tvmaze_ep_overview
    tvmaze_ep_runtime

    """

    # Series info
    series_map = {
        'tvmaze_series_name': 'name',
        'tvmaze_series_year': 'year',
        'tvdb_id': 'tvdb_id',
        'tvrage_id': 'tvrage_id',
        'tvmaze_series_id': 'tvmaze_id',
        'tvmaze_series_show_id': 'tvmaze_id',
        'tvmaze_series_tvrage': 'tvrage_id',
        'tvmaze_series_runtime': 'runtime',
        'tvmaze_series_premiered': 'premiered',
        'tvmaze_series_airdays': 'schedule',
        'tvmaze_series_weight': 'weight',
        'tvmaze_series_update_date': 'updated',
        'tvmaze_series_language': 'language',
        'tvmaze_series_original_image': 'original_image',
        'tvmaze_series_medium_image': 'medium_image',
        'tvmaze_series_summary': 'summary',
        'tvmaze_series_webchannel': 'webchannel',
        'tvmaze_series_show_type': 'show_type',
        'tvmaze_genres': lambda i: [db_genre.name for db_genre in i.genres],
        'tvmaze_series_network': 'network',
        'tvmaze_series_url': 'url',
        'tvmaze_series_status': 'status',
        'tvmaze_series_rating': 'rating',
        'tvmaze_series_episodes': lambda show: [episodes.title for episodes in show.episodes],
    }

    # Season Info
    season_map = {
        'tvmaze_season_id': 'tvmaze_id',
        'tvmaze_season_series_id': 'series_id',
        'tvmaze_season_number': 'number',
        'tvmaze_season_url': 'url',
        'tvmaze_season_name': 'name',
        'tvmaze_season_episode_order': 'episode_order',
        'tvmaze_season_premiere_date': 'airdate',
        'tvmaze_season_end_date': 'end_date',
        'tvmaze_season_network': 'network',
        'tvmaze_season_web_channel': 'web_channel',
        'tvmaze_season_image': 'image',
        'tvmaze_season_summary': 'summary',
    }

    # Episode info
    episode_map = {
        'tvmaze_episode_name': 'title',
        'tvmaze_episode_season': 'season_number',
        'tvmaze_episode_number': 'number',
        'tvmaze_episode_id': 'tvmaze_id',
        'tvmaze_episode_airdate': 'airdate',
        'tvmaze_episode_url': 'url',
        'tvmaze_episode_summary': 'summary',
        'tvmaze_episode_original_image': 'original_image',
        'tvmaze_episode_medium_image': 'medium_image',
        'tvmaze_episode_airstamp': 'airstamp',
        'tvmaze_episode_runtime': 'runtime',
    }

    schema = {'type': 'boolean'}

    @entry.register_lazy_lookup('tvmaze_series_lookup')
    def lazy_series_lookup(self, entry):
        """Does the lookup for this entry and populates the entry fields."""
        series_lookup = plugin.get('api_tvmaze', self).series_lookup
        with Session() as session:
            lookupargs = {
                'title': entry.get('series_name', eval_lazy=False),
                'year': entry.get('year', eval_lazy=False),
                'tvmaze_id': entry.get('tvmaze_id', eval_lazy=False),
                'tvdb_id': entry.get('tvdb_id', eval_lazy=False),
                'tvrage_id': entry.get('tvrage_idk', eval_lazy=False),
                'session': session,
            }
            try:
                series = series_lookup(**lookupargs)
            except LookupError as e:
                logger.debug(e)
            else:
                entry.update_using_map(self.series_map, series)
        return entry

    @entry.register_lazy_lookup('tvmaze_season_lookup')
    def lazy_season_lookup(self, entry):
        season_lookup = plugin.get('api_tvmaze', self).season_lookup
        with Session(expire_on_commit=False) as session:
            lookupargs = {
                'title': entry.get('series_name', eval_lazy=False),
                'year': entry.get('year', eval_lazy=False),
                'tvmaze_id': entry.get('tvmaze_id', eval_lazy=False),
                'tvdb_id': entry.get('tvdb_id', eval_lazy=False),
                'tvrage_id': entry.get('tvrage_id', eval_lazy=False),
                'series_season': entry.get('series_season', eval_lazy=False),
                'session': session,
            }
            try:
                season = season_lookup(**lookupargs)
            except LookupError as e:
                logger.debug(e)
            else:
                entry.update_using_map(self.season_map, season)
        return entry

    @entry.register_lazy_lookup('tvmaze_episode_lookup')
    def lazy_episode_lookup(self, entry):
        episode_lookup = plugin.get('api_tvmaze', self).episode_lookup
        with Session(expire_on_commit=False) as session:
            lookupargs = {
                'title': entry.get('series_name', eval_lazy=False),
                'year': entry.get('year', eval_lazy=False),
                'tvmaze_id': entry.get('tvmaze_id', eval_lazy=False),
                'tvdb_id': entry.get('tvdb_id', eval_lazy=False),
                'tvrage_id': entry.get('tvrage_id', eval_lazy=False),
                'series_season': entry.get('series_season', eval_lazy=False),
                'series_episode': entry.get('series_episode', eval_lazy=False),
                'series_date': entry.get('series_date', eval_lazy=False),
                'series_id_type': entry.get('series_id_type', eval_lazy=False),
                'session': session,
            }
            try:
                episode = episode_lookup(**lookupargs)
            except LookupError as e:
                logger.debug(e)
            else:
                entry.update_using_map(self.episode_map, episode)
        return entry

    # Run after series and metainfo series
    @plugin.priority(110)
    def on_task_metainfo(self, task, config):
        if not config:
            return

        for entry in task.entries:
            if (
                entry.get('series_name')
                or entry.get('tvdb_id', eval_lazy=False)
                or entry.get('tvmaze_id', eval_lazy=False)
                or entry.get('tvrage_id', eval_lazy=False)
            ):
                entry.add_lazy_fields(self.lazy_series_lookup, self.series_map)
                if entry.get('season_pack', eval_lazy=False):
                    entry.add_lazy_fields(self.lazy_season_lookup, self.season_map)
                if ('series_season' in entry and 'series_episode' in entry) or (
                    'series_date' in entry
                ):
                    entry.add_lazy_fields(self.lazy_episode_lookup, self.episode_map)

    @property
    def series_identifier(self):
        """Returns the plugin main identifier type"""
        return 'tvmaze_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        PluginTVMazeLookup, 'tvmaze_lookup', api_ver=2, interfaces=['task', 'series_metainfo']
    )
